# -*- coding: utf-8 -*-
#
#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Tagir Akhmetshin <tagirshin@gmail.com>
#  Copyright 2019 Dayana Bashirova <dayana.bashirova@yandex.ru>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import FrozenDict
from .element import Element
from .groups import GroupX
from .periods import PeriodIV, PeriodV, PeriodVI, PeriodVII


class Ni(Element, PeriodIV, GroupX):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 28

    @property
    def isotopes_distribution(self):
        return FrozenDict({58: 0.680769, 60: 0.262231, 61: 0.011399, 62: 0.036345, 64: 0.009256, 63: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({58: 57.935348, 60: 59.930791, 61: 60.93106, 62: 61.928349, 64: 63.92797, 63: 62.929669})

    @property
    def _common_valences(self):
        return 0, 2

    @property
    def _valences_exceptions(self):
        return ((2, False, 0, ()),
                (0, False, 0, ((2, 'O'), (1, 'O'))))  # Ni2O3

    @property
    def atomic_radius(self):
        return 1.49


class Pd(Element, PeriodV, GroupX):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 46

    @property
    def isotopes_distribution(self):
        return FrozenDict({102: 0.0102, 104: 0.1114, 105: 0.2233, 106: 0.2733, 108: 0.2646, 110: 0.1172, 103: 0.,
                           109: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({102: 101.905608, 104: 103.904035, 105: 104.905084, 106: 105.903483, 108: 107.903894,
                           110: 109.905152, 103: 102.906087, 109: 108.905950})

    @property
    def _common_valences(self):
        return 0, 2

    @property
    def _valences_exceptions(self):
        return ((2, False, 0, ()),
                (-2, False, 0, ((1, 'O'), (1, 'O'), (1, 'O'), (1, 'O'))),  # [Pd(OH)4]2-
                (-2, False, 0, ((1, 'C'), (1, 'C'), (1, 'C'), (1, 'C'))),  # [Pd(CN)4]2-
                (-2, False, 0, ((1, 'N'), (1, 'N'), (1, 'N'), (1, 'N'))),  # [Pd(NCS)4]2-
                (-2, False, 0, ((1, 'S'), (1, 'S'), (1, 'S'), (1, 'S'))),  # [Pd(SCN)4]2-
                (-2, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),  # [PdF4]2-
                (-2, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))))  # [PdCl4]2-

    @property
    def atomic_radius(self):
        return 1.69


class Pt(Element, PeriodVI, GroupX):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 78

    @property
    def isotopes_distribution(self):
        return FrozenDict({190: 0.00014, 192: 0.00782, 194: 0.32967, 195: 0.33832, 196: 0.25242, 198: 0.07163})

    @property
    def isotopes_masses(self):
        return FrozenDict({190: 189.95993, 192: 191.961035, 194: 193.962664, 195: 194.964774, 196: 195.964935,
                           198: 197.967876})

    @property
    def _common_valences(self):
        return 0, 2

    @property
    def _valences_exceptions(self):
        return ((0, False, 0, ((1, 'N'), (1, 'N'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),  # PtF6
                (0, False, 0, ((2, 'O'), (2, 'O'), (2, 'O'))))  # PtO3

    @property
    def atomic_radius(self):
        return 1.77


class Ds(Element, PeriodVII, GroupX):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 110

    @property
    def isotopes_distribution(self):
        return FrozenDict({281: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({281: 281.164516})

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ()

    @property
    def atomic_radius(self):
        return 1.77  # unknown, taken radius of previous element in group


__all__ = ['Ni', 'Pd', 'Pt', 'Ds']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict, namedtuple
from importlib.util import find_spec
from io import StringIO, BytesIO, TextIOWrapper, BufferedIOBase, BufferedReader
from itertools import count
from logging import warning
from pathlib import Path
from traceback import format_exc
from warnings import warn
from ._MDLrw import MDLStereo
from ..containers import MoleculeContainer, ReactionContainer
from ..exceptions import EmptyMolecule


parse_error = namedtuple('MRVParseError', ('number', 'json', 'log'))


def xml_dict(parent_element, stop_list=None):
    stop_list = set() if stop_list is None else set(stop_list)
    out = {}
    for x, y in parent_element.items():
        y = y.strip()
        if y:
            x = '@%s' % x.strip()
            out[x] = y

    text = []
    if len(parent_element):
        elements_grouped = defaultdict(list)
        for element in parent_element:
            name = QName(element).localname
            if name in stop_list:
                text.append(tostring(element, encoding=str, with_tail=False))
            else:
                elements_grouped[name].append(element)

            if element.tail:
                t = element.tail.strip()
                if t:
                    text.append(t)

        for element_tag, element_group in elements_grouped.items():
            if len(element_group) == 1:
                out[element_tag] = xml_dict(element_group[0], stop_list)
            else:
                out[element_tag] = [xml_dict(x, stop_list) for x in element_group]

    if parent_element.text:
        t = parent_element.text.strip()
        if t:
            text.insert(0, t)
    if text:
        out['$'] = ''.join(text)

    return out


class MRVRead(MDLStereo):
    """
    ChemAxon MRV files reader. works similar to opened file object. support `with` context manager.
    on initialization accept opened in binary mode file, string path to file,
    pathlib.Path object or another binary buffered reader object
    """
    def __init__(self, file, **kwargs):
        """
        :param ignore: Skip some checks of data or try to fix some errors.
        :param remap: Remap atom numbers started from one.
        :param store_log: Store parser log if exists messages to `.meta` by key `CGRtoolsParserLog`.
        :param calc_cis_trans: Calculate cis/trans marks from 2d coordinates.
        :param ignore_stereo: Ignore stereo data.
        """
        if isinstance(file, str):
            self.__file = open(file, 'rb')
            self.__is_buffer = False
        elif isinstance(file, Path):
            self.__file = file.open('rb')
            self.__is_buffer = False
        elif isinstance(file, (BytesIO, BufferedReader, BufferedIOBase)):
            self.__file = file
            self.__is_buffer = True
        else:
            raise TypeError('invalid file. BytesIO, BufferedReader and BufferedIOBase subclasses possible')
        super().__init__(**kwargs)
        self._data = self.__reader()

    def close(self, force=False):
        """
        close opened file

        :param force: force closing of externally opened file or buffer
        """
        if not self.__is_buffer or force:
            self.__file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def read(self):
        """
        parse whole file

        :return: list of parsed molecules or reactions
        """
        return list(iter(self))

    def __iter__(self):
        return (x for x in self._data if not isinstance(x, parse_error))

    def __next__(self):
        return next(iter(self))

    def __reader(self):
        for n, (_, element) in enumerate(iterparse(self.__file, tag='{*}MChemicalStruct')):
            parsed = xml_dict(element)
            element.clear()
            if 'molecule' in parsed and isinstance(parsed['molecule'], dict):
                parsed = parsed['molecule']
                try:
                    record = self.__parse_molecule(parsed)
                except (KeyError, ValueError):
                    self._info(f'record consist errors:\n{format_exc()}')
                    yield parse_error(n, parsed, self._format_log())
                else:
                    if 'propertyList' in parsed and 'property' in parsed['propertyList']:
                        record['meta'] = self.__parse_property(parsed['propertyList']['property'])
                    else:
                        record['meta'] = {}
                    try:
                        container = self._convert_structure(record)
                    except ValueError:
                        self._info(f'record consist errors:\n{format_exc()}')
                        yield parse_error(n, parsed, self._format_log())
                    else:
                        if self._store_log:
                            log = self._format_log()
                            if log:
                                container.meta['CGRtoolsParserLog'] = log
                        yield container
            elif 'reaction' in parsed and isinstance(parsed['reaction'], dict):
                parsed = parsed['reaction']
                try:
                    record = self.__parse_reaction(parsed)
                except (KeyError, ValueError):
                    self._info(f'record consist errors:\n{format_exc()}')
                    yield parse_error(n, parsed, self._format_log())
                else:
                    if 'propertyList' in parsed and 'property' in parsed['propertyList']:
                        record['meta'] = self.__parse_property(parsed['propertyList']['property'])
                    else:
                        record['meta'] = {}
                    try:
                        container = self._convert_reaction(record)
                    except ValueError:
                        self._info(f'record consist errors:\n{format_exc()}')
                        yield parse_error(n, parsed, self._format_log())
                    else:
                        if self._store_log:
                            log = self._format_log()
                            if log:
                                container.meta['CGRtoolsParserLog'] = log
                        yield container
            else:
                self._info('invalid MDocument')
                yield parse_error(n, parsed, self._format_log())
            self._flush_log()

    def __parse_reaction(self, data):
        reaction = {'reactants': [], 'products': [], 'reagents': []}
        title = data.get('@title')
        if title:
            reaction['title'] = title
        for tag, group in (('reactantList', 'reactants'), ('productList', 'products'), ('agentList', 'reagents')):
            if tag in data and 'molecule' in data[tag]:
                molecule = data[tag]['molecule']
                if isinstance(molecule, dict):
                    molecule = (molecule,)
                for m in molecule:
                    try:
                        reaction[group].append(self.__parse_molecule(m))
                    except EmptyMolecule:
                        if not self._ignore:
                            raise
                        self._info('empty molecule ignored')
        return reaction

    def __parse_property(self, data):
        meta = {}
        if isinstance(data, dict):
            key = data['@title']
            val = data['scalar']['$'].strip()
            if key and val:
                meta[key] = val
            else:
                self._info(f'invalid metadata entry: {data}')
        else:
            for x in data:
                key = x['@title']
                val = x['scalar']['$'].strip()
                if key and val:
                    meta[key] = val
                else:
                    self._info(f'invalid metadata entry: {x}')
        return meta

    def __parse_molecule(self, data):
        atoms, bonds, stereo = [], [], []
        atom_map = {}
        if 'atom' in data['atomArray']:
            da = data['atomArray']['atom']
            if isinstance(da, dict):
                da = (da,)
            for n, atom in enumerate(da):
                atom_map[atom['@id']] = n
                atoms.append({'element': atom['@elementType'],
                              'isotope': int(atom['@isotope']) if '@isotope' in atom else None,
                              'charge': int(atom.get('@formalCharge', 0)),
                              'is_radical': '@radical' in atom,
                              'mapping': int(atom.get('@mrvMap', 0))})
                if '@z3' in atom:
                    atoms[-1].update(x=float(atom['@x3']), y=float(atom['@y3']), z=float(atom['@z3']))
                else:
                    atoms[-1].update(x=float(atom['@x2']) / 2, y=float(atom['@y2']) / 2, z=0.)
                if '@mrvQueryProps' in atom:
                    raise ValueError('queries unsupported')
        else:
            atom = data['atomArray']
            for n, (_id, e) in enumerate(zip(atom['@atomID'].split(), atom['@elementType'].split())):
                atom_map[_id] = n
                atoms.append({'element': e, 'charge': 0, 'mapping': 0, 'isotope': None, 'is_radical': False})
            if '@z3' in atom:
                for a, x, y, z in zip(atoms, atom['@x3'].split(), atom['@y3'].split(), atom['@z3'].split()):
                    a['x'] = float(x)
                    a['y'] = float(y)
                    a['z'] = float(z)
            else:
                for a, x, y in zip(atoms, atom['@x2'].split(), atom['@y2'].split()):
                    a['x'] = float(x) / 2
                    a['y'] = float(y) / 2
                    a['z'] = 0.
            if '@isotope' in atom:
                for a, x in zip(atoms, atom['@isotope'].split()):
                    if x != '0':
                        a['isotope'] = int(x)
            if '@formalCharge' in atom:
                for a, x in zip(atoms, atom['@formalCharge'].split()):
                    if x != '0':
                        a['charge'] = int(x)
            if '@mrvMap' in atom:
                for a, x in zip(atoms, atom['@mrvMap'].split()):
                    if x != '0':
                        a['mapping'] = int(x)
            if '@radical' in atom:
                for a, x in zip(atoms, atom['@radical'].split()):
                    if x != '0':
                        a['is_radical'] = True
            if '@mrvQueryProps' in atom:
                raise ValueError('queries unsupported')
        if not atoms:
            raise EmptyMolecule

        if 'bond' in data['bondArray']:
            db = data['bondArray']['bond']
            if isinstance(db, dict):
                db = (db,)
            for bond in db:
                order = self.__bond_map[bond['@queryType' if '@queryType' in bond else '@order']]
                a1, a2 = bond['@atomRefs2'].split()
                if 'bondStereo' in bond:
                    if '$' in bond['bondStereo']:
                        s = bond['bondStereo']['$']
                        if s == 'H':
                            stereo.append((atom_map[a1], atom_map[a2], -1))
                        elif s == 'W':
                            stereo.append((atom_map[a1], atom_map[a2], 1))
                        else:
                            self._info('invalid or unsupported stereo')
                    else:
                        self._info('incorrect bondStereo tag')
                bonds.append((atom_map[a1], atom_map[a2], order))

        mol = {'atoms': atoms, 'bonds': bonds, 'stereo': stereo}
        if '@title' in data:
            mol['title'] = data['@title']
        return mol

    __bond_map = {'Any': 8, 'any': 8, 'A': 4, 'a': 4, '1': 1, '2': 2, '3': 3}
    __radical_map = {'monovalent': 2, 'divalent': 1, 'divalent1': 1, 'divalent3': 3}


class MRVWrite:
    """
    ChemAxon MRV files writer. works similar to opened for writing file object. support `with` context manager.
    on initialization accept opened for writing in text mode file, string path to file,
    pathlib.Path object or another buffered writer object
    """
    def __init__(self, file):
        if isinstance(file, str):
            self._file = open(file, 'w')
            self._is_buffer = False
        elif isinstance(file, Path):
            self._file = file.open('w')
            self._is_buffer = False
        elif isinstance(file, (TextIOWrapper, StringIO)):
            self._file = file
            self._is_buffer = True
        else:
            raise TypeError('invalid file. '
                            'TextIOWrapper, StringIO, BytesIO, BufferedReader and BufferedIOBase subclasses possible')
        self.__writable = True

    def close(self, force=False):
        """
        write close tag of MRV file and close opened file

        :param force: force closing of externally opened file or buffer
        """
        if not self.__finalized:
            self._file.write('</cml>\n')
            self.__finalized = True
        if self.__writable:
            self.write = self.__write_closed
            self.__writable = False

        if not self._is_buffer or force:
            self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    @staticmethod
    def __write_closed(_):
        raise ValueError('I/O operation on closed writer')

    def write(self, data):
        """
        write single molecule or reaction into file
        """
        self._file.write('<cml>\n')
        self.__write(data)
        self.write = self.__write

    def __write(self, data):
        if isinstance(data, ReactionContainer):
            buffer = ['<MDocument><MChemicalStruct>']
            if not data._arrow:
                data.fix_positions()

            if data.name:
                buffer.append(f'<reaction title="{data.name}">')
            else:
                buffer.append('<reaction>')

            if data.meta:
                buffer.append('<propertyList>')
                for k, v in data.meta.items():
                    if isinstance(v, str):
                        v = f'<![CDATA[{v}]]>'
                    buffer.append(f'<property title="{k}"><scalar>{v}</scalar></property>')
                buffer.append('</propertyList>')
            c = count(1)
            for i, j in ((data.reactants, 'reactantList'), (data.products, 'productList'),
                         (data.reagents, 'agentList')):
                if not i:
                    continue
                buffer.append(f'<{j}>')
                for n, m in zip(c, i):
                    if m.name:
                        buffer.append(f'<molecule title="{m.name}" molID="m{n}">')
                    else:
                        buffer.append(f'<molecule molID="m{n}">')
                    buffer.append(self.__convert_structure(m))
                    buffer.append('</molecule>')
                buffer.append(f'</{j}>')

            buffer.append(f'<arrow type="DEFAULT" x1="{data._arrow[0] * 2:.4f}" y1="0" '
                          f'x2="{data._arrow[1] * 2:.4f}" y2="0"/>')
            buffer.append('</reaction>')
            self._file.writelines(buffer)
        else:
            m = self.__convert_structure(data)
            self._file.write('<MDocument><MChemicalStruct>')

            if data.name:
                self._file.write(f'<molecule title="{data.name}">')
            else:
                self._file.write('<molecule>')

            if data.meta:
                self._file.write('<propertyList>')
                for k, v in data.meta.items():
                    if isinstance(v, str):
                        v = f'<![CDATA[{v}]]>'
                    self._file.write(f'<property title="{k}"><scalar>{v}</scalar></property>')
                self._file.write('</propertyList>')
            self._file.write(m)
            self._file.write('</molecule>')
        self._file.write('</MChemicalStruct></MDocument>\n')

    def __convert_structure(self, g):
        if not isinstance(g, MoleculeContainer):
            raise TypeError('MoleculeContainer expected')
        bond_map = self.__bond_map
        gp = g._plane
        gc = g._charges
        gr = g._radicals
        bg = g._bonds

        out = ['<atomArray>']
        for n, atom in g._atoms.items():
            x, y = gp[n]
            out.append(f'<atom id="a{n}" elementType="{atom.atomic_symbol}" '
                       f'x2="{x * 2:.4f}" y2="{y * 2:.4f}" mrvMap="{n}"')
            if gc[n]:
                out.append(f' formalCharge="{gc[n]}"')
            if gr[n]:
                out.append(' radical="monovalent"')
            if atom.isotope:
                out.append(f' isotope="{atom.isotope}"')
            out.append('/>')
        out.append('</atomArray>')

        out.append('<bondArray>')
        wedge = defaultdict(set)
        for n, (i, j, s) in enumerate(g._wedge_map, start=1):
            out.append(f'<bond id="b{n}" atomRefs2="a{i} a{j}" order="{bond_map[bg[i][j].order]}">'
                       f'<bondStereo>{s == 1 and "W" or "H"}</bondStereo></bond>')
            wedge[i].add(j)
            wedge[j].add(i)
        for n, (i, j, bond) in enumerate(g.bonds(), start=len(out)):
            if j not in wedge[i]:
                out.append(f'<bond id="b{n}" atomRefs2="a{i} a{j}" order="{bond_map[bond.order]}"/>')
        out.append('</bondArray>')
        return ''.join(out)

    __bond_map = {8: '1" queryType="Any', 4: 'A', 1: '1', 2: '2', 3: '3', None: '0'}
    __finalized = False


class MRVread:
    def __init__(self, *args, **kwargs):
        warn('MRVread deprecated. Use MRVRead instead', DeprecationWarning)
        warning('MRVread deprecated. Use MRVRead instead')
        self.__obj = MRVRead(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self.__obj, item)

    def __iter__(self):
        return iter(self.__obj)

    def __next__(self):
        return next(self.__obj)

    def __enter__(self):
        return self.__obj.__enter__()

    def __exit__(self, _type, value, traceback):
        return self.__obj.__exit__(_type, value, traceback)


class MRVwrite:
    def __init__(self, *args, **kwargs):
        warn('MRVwrite deprecated. Use MRVWrite instead', DeprecationWarning)
        warning('MRVwrite deprecated. Use MRVWrite instead')
        self.__obj = MRVWrite(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self.__obj, item)

    def __enter__(self):
        return self.__obj.__enter__()

    def __exit__(self, _type, value, traceback):
        return self.__obj.__exit__(_type, value, traceback)


__all__ = ['MRVWrite', 'MRVwrite']

if find_spec('lxml'):
    from lxml.etree import iterparse, QName, tostring

    __all__.extend(['MRVRead', 'MRVread'])
else:
    warn('lxml library not installed', ImportWarning)
    del MRVRead, MRVread



# -*- coding: utf-8 -*-
#
#  Copyright 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import Type
from .core import Core
from ...exceptions import IsNotConnectedAtom


class Dynamic(Core):
    __slots__ = ()

    @Core.charge.setter
    def charge(self, charge):
        try:
            g = self._graph()
            g._charges[self._map] = g._validate_charge(charge)
            g.flush_cache()
        except AttributeError:
            raise IsNotConnectedAtom

    @Core.is_radical.setter
    def is_radical(self, is_radical):
        try:
            g = self._graph()
            g._radicals[self._map] = g._validate_radical(is_radical)
            g.flush_cache()
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def p_charge(self) -> int:
        try:
            return self._graph()._p_charges[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @p_charge.setter
    def p_charge(self, charge):
        try:
            g = self._graph()
            g._p_charges[self._map] = g._validate_charge(charge)
            g.flush_cache()
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def p_is_radical(self) -> bool:
        try:
            return self._graph()._p_radicals[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @p_is_radical.setter
    def p_is_radical(self, is_radical):
        try:
            g = self._graph()
            g._p_radicals[self._map] = g._validate_radical(is_radical)
            g.flush_cache()
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def p_hybridization(self):
        """
        Product state hybridization of atom
        """
        try:
            return self._graph()._p_hybridizations[self._map]
        except AttributeError:
            raise IsNotConnectedAtom


class DynamicElement(Dynamic):
    __slots__ = ('__p_charge', '__p_is_radical')

    @property
    def atomic_symbol(self) -> str:
        return self.__class__.__name__[7:]

    @classmethod
    def from_symbol(cls, symbol: str) -> Type['DynamicElement']:
        """
        get DynamicElement class by its symbol
        """
        try:
            element = next(x for x in DynamicElement.__subclasses__() if x.__name__ == f'Dynamic{symbol}')
        except StopIteration:
            raise ValueError(f'DynamicElement with symbol "{symbol}" not found')
        return element

    @classmethod
    def from_atomic_number(cls, number: int) -> Type['DynamicElement']:
        """
        get DynamicElement class by its number
        """
        try:
            element = next(x for x in DynamicElement.__subclasses__() if x.atomic_number.fget(None) == number)
        except StopIteration:
            raise ValueError(f'DynamicElement with number "{number}" not found')
        return element

    @property
    def neighbors(self):
        try:
            return sum(b.order is not None for b in self._graph()._bonds[self._map].values())
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def p_neighbors(self):
        try:
            return sum(b.p_order is not None for b in self._graph()._bonds[self._map].values())
        except AttributeError:
            raise IsNotConnectedAtom

    def __eq__(self, other):
        """
        compare attached to molecules dynamic elements
        """
        return isinstance(other, DynamicElement) and self.atomic_number == other.atomic_number and \
            self.isotope == other.isotope and self.charge == other.charge and self.is_radical == other.is_radical and \
            self.p_charge == other.p_charge and self.p_is_radical == other.p_is_radical

    def __hash__(self):
        """
        26bit = 9bit | 7bit | 4bit | 4bit | 1bit| 1bit
        """
        return (self.isotope or 0) << 17 | self.atomic_number << 10 | self.charge + 4 << 6 | \
            self.p_charge + 4 << 2 | self.is_radical << 1 | self.p_is_radical


__all__ = ['DynamicElement', 'Dynamic']



# -*- coding: utf-8 -*-
#
#  Copyright 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod
from typing import Optional, Tuple, Dict
from weakref import ref
from ...exceptions import IsConnectedAtom, IsNotConnectedAtom


class Core(ABC):
    __slots__ = ('__isotope', '_graph', '_map', '_backward')

    def __init__(self, isotope: Optional[int] = None):
        """
        Element object with specified isotope

        :param isotope: Isotope number of element
        """
        if isinstance(isotope, int):
            if isotope not in self.isotopes_distribution:
                raise ValueError(f'isotope number {isotope} impossible or not stable for {self.__class__.__name__}')
        elif isotope is not None:
            raise TypeError('integer isotope number required')
        self.__isotope = isotope

    def __repr__(self):
        if self.__isotope:
            return f'{self.__class__.__name__}({self.__isotope})'
        return f'{self.__class__.__name__}()'

    def __getstate__(self):
        return {'isotope': self.__isotope}

    def __setstate__(self, state):
        self.__isotope = state['isotope']

    def __int__(self):
        """
        Same as hash
        """
        return hash(self)

    @property
    @abstractmethod
    def atomic_number(self) -> int:
        """
        Element number
        """

    @property
    def isotope(self) -> Optional[int]:
        """
        Isotope number
        """
        return self.__isotope

    @property
    def atomic_mass(self) -> float:
        mass = self.isotopes_masses
        if self.__isotope is None:
            return sum(x * mass[i] for i, x in self.isotopes_distribution.items())
        return mass[self.__isotope]

    @property
    @abstractmethod
    def isotopes_distribution(self) -> Dict[int, float]:
        """
        Isotopes distribution in earth
        """

    @property
    @abstractmethod
    def isotopes_masses(self) -> Dict[int, float]:
        """
        Isotopes masses
        """

    @property
    @abstractmethod
    def atomic_radius(self) -> float:
        """
        Valence radius of atom
        """

    @property
    def charge(self) -> int:
        """
        Charge of atom
        """
        try:
            return self._graph()._charges[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def is_radical(self) -> bool:
        """
        Radical state of atoms
        """
        try:
            return self._graph()._radicals[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def x(self) -> float:
        """
        X coordinate of atom on 2D plane
        """
        try:
            return self._graph()._plane[self._map][0]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def y(self) -> float:
        """
        Y coordinate of atom on 2D plane
        """
        try:
            return self._graph()._plane[self._map][1]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def xy(self) -> Tuple[float, float]:
        """
        (X, Y) coordinates of atom on 2D plane
        """
        try:
            return self._graph()._plane[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    @abstractmethod
    def neighbors(self):
        """
        Neighbors count of atom
        """

    @property
    def hybridization(self):
        """
        1 - if atom has zero or only single bonded neighbors, 2 - if has only one double bonded neighbor and any amount
        of single bonded, 3 - if has one triple bonded and any amount of double and single bonded neighbors or
        two double bonded and any amount of single bonded neighbors, 4 - if atom in aromatic ring.
        """
        try:
            return self._graph()._hybridizations[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def in_ring(self) -> bool:
        """
        Atom in any ring.
        """
        try:
            return self._map in self._graph().ring_atoms
        except AttributeError:
            raise IsNotConnectedAtom

    def copy(self) -> 'Core':
        """
        Detached from graph copy of element
        """
        copy = object.__new__(self.__class__)
        copy._Core__isotope = self.__isotope
        return copy

    def _attach_to_graph(self, graph, _map):
        try:
            self._graph
        except AttributeError:
            self._graph = ref(graph)
            self._map = _map
        else:
            raise IsConnectedAtom

    def _change_map(self, _map):
        try:
            self._graph
        except AttributeError:
            raise IsNotConnectedAtom
        else:
            self._map = _map


__all__ = ['Core']



# -*- coding: utf-8 -*-
#
#  Copyright 2014-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from bisect import bisect_left
from collections import defaultdict
from io import BytesIO
from logging import warning
from subprocess import check_output
from traceback import format_exc
from warnings import warn
from ._CGRrw import parse_error
from ._MDLrw import MDLRead, MDLWrite, MOLRead, EMOLRead


class SDFRead(MDLRead):
    """
    MDL SDF files reader. works similar to opened file object. support `with` context manager.
    on initialization accept opened in text mode file, string path to file,
    pathlib.Path object or another buffered reader object
    """
    def __init__(self, file, indexable=False, **kwargs):
        """
        :param indexable: if True: supported methods seek, tell, object size and subscription, it only works when
            dealing with a real file (the path to the file is specified) because the external grep utility is used,
            supporting in unix-like OS the object behaves like a normal open file.

            if False: works like generator converting a record into MoleculeContainer and returning each object in
            order, records with errors are skipped
        :param ignore: Skip some checks of data or try to fix some errors.
        :param remap: Remap atom numbers started from one.
        :param store_log: Store parser log if exists messages to `.meta` by key `CGRtoolsParserLog`.
        :param calc_cis_trans: Calculate cis/trans marks from 2d coordinates.
        :param ignore_stereo: Ignore stereo data.
        """
        super().__init__(file, **kwargs)
        self.__file = iter(self._file.readline, '')
        self._data = self.__reader()
        next(self._data)

        if indexable:
            self._load_cache()

    @staticmethod
    def _get_shifts(file):
        shifts = [0]
        for x in BytesIO(check_output(['grep', '-bE', r'\$\$\$\$', file])):
            pos, line = x.split(b':', 1)
            shifts.append(int(pos) + len(line))
        return shifts

    def seek(self, offset):
        """
        shifts on a given number of record in the original file
        :param offset: number of record
        """
        if self._shifts:
            if 0 <= offset < len(self._shifts):
                current_pos = self._file.tell()
                new_pos = self._shifts[offset]
                if current_pos != new_pos:
                    if current_pos == self._shifts[-1]:  # reached the end of the file
                        self._file.seek(new_pos)
                        self.__file = iter(self._file.readline, '')
                        self._data = self.__reader()
                        next(self._data)
                        if offset:
                            self._data.send(offset)
                            self.__already_seeked = True
                    elif not self.__already_seeked:
                        self._file.seek(new_pos)
                        self._data.send(offset)
                        self.__already_seeked = True
                    else:
                        raise BlockingIOError('File already seeked. New seek possible only after reading any data')
            else:
                raise IndexError('invalid offset')
        else:
            raise self._implement_error

    def tell(self):
        """
        :return: number of records processed from the original file
        """
        if self._shifts:
            t = self._file.tell()
            return bisect_left(self._shifts, t)
        raise self._implement_error

    def __reader(self):
        im = 3
        failkey = False
        mkey = parser = record = None
        meta = defaultdict(list)
        file = self._file
        seekable = file.seekable()
        seek = yield  # init stop
        if seek is not None:
            yield
            pos = file.tell()
            count = seek
            self.__already_seeked = False
        else:
            pos = 0 if seekable else None
            count = 0
        for line in self.__file:
            if failkey and not line.startswith("$$$$"):
                continue
            elif parser:
                try:
                    if parser(line):
                        record = parser.getvalue()
                        parser = None
                except ValueError:
                    parser = None
                    self._info(f'line:\n{line}\nconsist errors:\n{format_exc()}')
                    seek = yield parse_error(count, pos, self._format_log())
                    if seek is not None:  # seeked to start of mol block
                        yield
                        count = seek
                        pos = file.tell()
                        im = 3
                        mkey = None
                        meta = defaultdict(list)
                        self.__already_seeked = False
                    else:
                        failkey = True
                    self._flush_log()
            elif line.startswith("$$$$"):
                if record:
                    record['meta'] = self._prepare_meta(meta)
                    if title:
                        record['title'] = title
                    try:
                        container = self._convert_structure(record)
                    except ValueError:
                        self._info(f'record consist errors:\n{format_exc()}')
                        seek = yield parse_error(count, pos, self._format_log())
                    else:
                        if self._store_log:
                            log = self._format_log()
                            if log:
                                container.meta['CGRtoolsParserLog'] = log
                        seek = yield container
                    if seek is not None:  # seeked position
                        yield
                        count = seek - 1
                        self.__already_seeked = False
                    self._flush_log()
                    record = None

                if seekable:
                    pos = file.tell()
                count += 1
                im = 3
                failkey = False
                mkey = None
                meta = defaultdict(list)
            elif record:
                if line.startswith('>  <'):
                    mkey = line.rstrip()[4:-1].strip()
                    if not mkey:
                        self._info(f'invalid metadata entry: {line}')
                elif mkey:
                    data = line.strip()
                    if data:
                        meta[mkey].append(data)
            elif im:
                if im == 3:  # parse mol title
                    title = line.strip()
                im -= 1
            elif not im:
                try:
                    if 'V2000' in line:
                        parser = MOLRead(line, self._log_buffer)
                    elif 'V3000' in line:
                        parser = EMOLRead(self._log_buffer)
                    else:
                        raise ValueError('invalid MOL entry')
                except ValueError:
                    self._info(f'line:\n{line}\nconsist errors:\n{format_exc()}')
                    seek = yield parse_error(count, pos, self._format_log())
                    if seek is not None:  # seeked to start of mol block
                        yield
                        count = seek
                        pos = file.tell()
                        im = 3
                        mkey = None
                        meta = defaultdict(list)
                        self.__already_seeked = False
                    else:
                        failkey = True
                    self._flush_log()

        if record:  # True for MOL file only.
            record['meta'] = self._prepare_meta(meta)
            if title:
                record['title'] = title
            try:
                container = self._convert_structure(record)
            except ValueError:
                self._info(f'record consist errors:\n{format_exc()}')
                log = self._format_log()
                self._flush_log()
                yield parse_error(count, pos, log)
            else:
                if self._store_log:
                    log = self._format_log()
                    if log:
                        container.meta['CGRtoolsParserLog'] = log
                self._flush_log()
                yield container

    __already_seeked = False


class SDFWrite(MDLWrite):
    """
    MDL SDF files writer. works similar to opened for writing file object. support `with` context manager.
    on initialization accept opened for writing in text mode file, string path to file,
    pathlib.Path object or another buffered writer object
    """
    def write(self, data):
        """
        write single molecule into file
        """
        mol = self._convert_structure(data)
        if isinstance(mol, list):
            self._file.write('$$$$\n'.join(mol))
        else:
            self._file.write(mol)

        for k, v in data.meta.items():
            self._file.write(f'>  <{k}>\n{v}\n')
        self._file.write('$$$$\n')


class SDFread:
    def __init__(self, *args, **kwargs):
        warn('SDFread deprecated. Use SDFRead instead', DeprecationWarning)
        warning('SDFread deprecated. Use SDFRead instead')
        self.__obj = SDFRead(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self.__obj, item)

    def __iter__(self):
        return iter(self.__obj)

    def __next__(self):
        return next(self.__obj)

    def __getitem__(self, item):
        return self.__obj[item]

    def __enter__(self):
        return self.__obj.__enter__()

    def __exit__(self, _type, value, traceback):
        return self.__obj.__exit__(_type, value, traceback)

    def __len__(self):
        return len(self.__obj)


class SDFwrite:
    def __init__(self, *args, **kwargs):
        warn('SDFwrite deprecated. Use SDFWrite instead', DeprecationWarning)
        warning('SDFwrite deprecated. Use SDFWrite instead')
        self.__obj = SDFWrite(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self.__obj, item)

    def __enter__(self):
        return self.__obj.__enter__()

    def __exit__(self, _type, value, traceback):
        return self.__obj.__exit__(_type, value, traceback)


__all__ = ['SDFRead', 'SDFWrite', 'SDFread', 'SDFwrite']



# -*- coding: utf-8 -*-
#
#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import Optional


class Bond:
    __slots__ = ('__order',)

    def __init__(self, order):
        if not isinstance(order, int):
            raise TypeError('invalid order value')
        if order not in (1, 4, 2, 3, 8):
            raise ValueError('order should be from [1, 2, 3, 4, 8]')
        self.__order = order

    def __eq__(self, other):
        if isinstance(other, Bond):
            return self.__order == other.order
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__order})'

    def __int__(self):
        return self.__order

    def __hash__(self):
        return self.__order

    @property
    def order(self) -> int:
        return self.__order

    def copy(self) -> 'Bond':
        copy = object.__new__(self.__class__)
        copy._Bond__order = self.__order
        return copy


class DynamicBond:
    __slots__ = ('__order', '__p_order')

    def __init__(self, order=None, p_order=None):
        if order is None:
            if not isinstance(p_order, int):
                raise TypeError('p_order should be int type')
        elif not isinstance(order, int):
            raise TypeError('order should be int type or None')
        elif p_order is not None and not isinstance(p_order, int):
            raise TypeError('p_order should be int type or None')

        if order not in (1, 4, 2, 3, None, 8) or p_order not in (1, 4, 2, 3, None, 8):
            raise ValueError('order or p_order should be from [1, 2, 3, 4, 8]')

        self.__order = order
        self.__p_order = p_order

    def __eq__(self, other):
        if isinstance(other, DynamicBond):
            return self.__order == other.order and self.__p_order == other.p_order
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__order}, {self.__p_order})'

    def __int__(self):
        return hash(self)

    def __hash__(self):
        return (self.__order or 0) * 10 + (self.__p_order or 0)

    @property
    def order(self) -> Optional[int]:
        return self.__order

    @property
    def p_order(self) -> Optional[int]:
        return self.__p_order

    def copy(self) -> 'DynamicBond':
        copy = object.__new__(self.__class__)
        copy._DynamicBond__order = self.__order
        copy._DynamicBond__p_order = self.__p_order
        return copy



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import cached_method
from collections.abc import Iterable
from functools import reduce
from hashlib import sha512
from itertools import chain
from operator import or_
from typing import Dict, Iterable as TIterable, Iterator, Optional, Tuple, Union
from .cgr import CGRContainer
from .cgr_query import QueryCGRContainer
from .molecule import MoleculeContainer
from .query import QueryContainer
from ..algorithms.components import ReactionComponents
from ..algorithms.depict import DepictReaction
from ..algorithms.standardize import StandardizeReaction


graphs = Union[MoleculeContainer, QueryContainer, CGRContainer, QueryCGRContainer]


class ReactionContainer(StandardizeReaction, ReactionComponents, DepictReaction):
    """
    Reaction storage. Contains reactants, products and reagents lists.

    Reaction storage hashable and comparable. based on reaction unique signature (SMIRKS).
    """
    __slots__ = ('__reactants', '__products', '__reagents', '__meta', '__name', '_arrow', '_signs', '__dict__')

    def __init__(self, reactants: TIterable[graphs] = (), products: TIterable[graphs] = (),
                 reagents: TIterable[graphs] = (), meta: Optional[Dict] = None, name: Optional[str] = None):
        """
        New reaction object creation

        :param reactants: list of MoleculeContainers [or other Structure Containers] in left side of reaction
        :param products: right side of reaction. see reactants
        :param reagents: middle side of reaction: solvents, catalysts, etc. see reactants
        :param meta: dictionary of metadata. like DTYPE-DATUM in RDF

        """
        if not isinstance(reactants, Iterable) or isinstance(reactants, (str, bytes)) or \
                not isinstance(products, Iterable) or isinstance(products, (str, bytes)) or \
                not isinstance(reagents, Iterable) or isinstance(reagents, (str, bytes)):
            raise TypeError('Iterator of molecules or CGRs or queries expected')
        reactants = tuple(reactants)
        products = tuple(products)
        reagents = tuple(reagents)
        try:
            base_type = next(chain(reactants, products, reagents)).__class__
        except StopIteration:
            raise ValueError('At least one graph object required')
        if not any(issubclass(base_type, x) for x in
                   (MoleculeContainer, QueryContainer, CGRContainer, QueryCGRContainer)):
            raise TypeError('MoleculeContainer or QueryContainer or CGRContainer or QueryCGRContainer expected')
        elif not all(isinstance(x, base_type) for x in chain(reactants, products, reagents)):
            raise TypeError(f'{base_type.__name__} expected for all graphs')

        self.__reactants = reactants
        self.__products = products
        self.__reagents = reagents
        if meta is None:
            self.__meta = {}
        else:
            self.__meta = dict(meta)
        if name is None:
            self.__name = ''
        else:
            self.name = name
        self._arrow = None
        self._signs = None

    @classmethod
    def from_cgr(cls, cgr: CGRContainer) -> 'ReactionContainer':
        """
        Decompose CGR into reaction
        """
        if not isinstance(cgr, CGRContainer):
            raise TypeError('CGR expected')
        r, p = ~cgr
        reaction = object.__new__(cls)
        reaction._ReactionContainer__reactants = tuple(r.split())
        reaction._ReactionContainer__products = tuple(p.split())
        reaction._ReactionContainer__reagents = ()
        reaction._ReactionContainer__meta = cgr._Graph__meta.copy()
        reaction._ReactionContainer__name = cgr._Graph__name
        reaction._arrow = None
        reaction._signs = None
        return reaction

    def __getitem__(self, item):
        if item in ('reactants', 0):
            return self.__reactants
        elif item in ('products', 1):
            return self.__products
        elif item in ('reagents', 2):
            return self.__reagents
        elif item == 'meta':
            return self.__meta
        elif item == 'name':
            return self.__name
        raise KeyError('invalid attribute')

    def __getstate__(self):
        return dict(reactants=self.__reactants, products=self.__products, reagents=self.__reagents, meta=self.__meta,
                    name=self.__name)

    def __setstate__(self, state):
        if next(iter(state)) == 'reagents':  # 3.0 compatibility
            state['reagents'], state['reactants'] = state['reactants'], state['reagents']
        self.__reactants = state['reactants']
        self.__products = state['products']
        self.__reagents = state['reagents']
        self.__meta = state['meta']
        self.__name = state.get('name', '')  # 4.0.9 compatibility
        self._arrow = None
        self._signs = None

    @property
    def reactants(self) -> Tuple[graphs, ...]:
        return self.__reactants

    @property
    def reagents(self) -> Tuple[graphs, ...]:
        return self.__reagents

    @property
    def products(self) -> Tuple[graphs, ...]:
        return self.__products

    def molecules(self) -> Iterator[graphs]:
        """
        Iterator of all reaction molecules
        """
        return chain(self.__reactants, self.__reagents, self.__products)

    @property
    def meta(self) -> Dict:
        """
        Dictionary of metadata.
        Like DTYPE-DATUM in RDF
        """
        return self.__meta

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str):
        if not isinstance(name, str):
            raise TypeError('name should be string up to 80 symbols')
        if len(name) > 80:
            raise ValueError('name should be string up to 80 symbols')
        self.__name = name

    def copy(self) -> 'ReactionContainer':
        """
        Get copy of object
        """
        copy = object.__new__(self.__class__)
        copy._ReactionContainer__reactants = tuple(x.copy() for x in self.__reactants)
        copy._ReactionContainer__products = tuple(x.copy() for x in self.__products)
        copy._ReactionContainer__reagents = tuple(x.copy() for x in self.__reagents)
        copy._ReactionContainer__meta = self.__meta.copy()
        copy._ReactionContainer__name = self.__name
        copy._arrow = self._arrow
        copy._signs = self._signs
        return copy

    @cached_method
    def compose(self) -> CGRContainer:
        """
        Get CGR of reaction

        Reagents will be presented as unchanged molecules
        :return: CGRContainer
        """
        rr = self.__reagents + self.__reactants
        if rr:
            if not isinstance(rr[0], (MoleculeContainer, CGRContainer)):
                raise TypeError('Queries not composable')
            r = reduce(or_, rr)
        else:
            r = MoleculeContainer()
        if self.__products:
            if not isinstance(self.__products[0], (MoleculeContainer, CGRContainer)):
                raise TypeError('Queries not composable')
            p = reduce(or_, self.__products)
        else:
            p = MoleculeContainer()
        c = r ^ p
        c.meta.update(self.__meta)
        return c

    def __invert__(self):
        """
        Get CGR of reaction
        """
        return self.compose()

    def __eq__(self, other):
        return isinstance(other, ReactionContainer) and str(self) == str(other)

    @cached_method
    def __hash__(self):
        return hash(str(self))

    @cached_method
    def __bytes__(self):
        return sha512(str(self).encode()).digest()

    def __bool__(self):
        return bool(self.__reactants or self.__products or self.__reagents)

    @cached_method
    def __str__(self):
        sig = []
        for ml in (self.__reactants, self.__reagents, self.__products):
            sig.append('.'.join(sorted(str(x) for x in ml)))
        return '>'.join(sig)

    def __format__(self, format_spec):
        sig = []
        for ml in (self.__reactants, self.__reagents, self.__products):
            sig.append('.'.join(sorted(format(x, format_spec) for x in ml)))
        return '>'.join(sig)

    def flush_cache(self):
        self.__dict__.clear()
        for m in self.molecules():
            m.flush_cache()


__all__ = ['ReactionContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import abstractmethod
from CachedMethods import class_cached_property
from collections import defaultdict
from typing import Optional, Tuple, Dict, Set, List, Type
from .core import Core
from ...exceptions import IsNotConnectedAtom, ValenceError


class Element(Core):
    __slots__ = ()
    __class_cache__ = {}

    @property
    def atomic_symbol(self) -> str:
        return self.__class__.__name__

    @Core.charge.setter
    def charge(self, charge: int):
        try:
            g = self._graph()
            g._charges[self._map] = g._validate_charge(charge)
            g._calc_implicit(self._map)
            g.flush_cache()
            g._fix_stereo()
        except AttributeError:
            raise IsNotConnectedAtom

    @Core.is_radical.setter
    def is_radical(self, is_radical: bool):
        try:
            g = self._graph()
            g._radicals[self._map] = g._validate_radical(is_radical)
            g._calc_implicit(self._map)
            g.flush_cache()
            g._fix_stereo()
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def implicit_hydrogens(self) -> Optional[int]:
        try:
            return self._graph()._hydrogens[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def explicit_hydrogens(self) -> int:
        try:
            return self._graph()._explicit_hydrogens(self._map)
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def total_hydrogens(self) -> int:
        try:
            return self._graph()._total_hydrogens(self._map)
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def neighbors(self) -> int:
        try:
            return len(self._graph()._bonds[self._map])
        except AttributeError:
            raise IsNotConnectedAtom

    @classmethod
    def from_symbol(cls, symbol: str) -> Type['Element']:
        """
        get Element class by its symbol
        """
        try:
            element = next(x for x in Element.__subclasses__() if x.__name__ == symbol)
        except StopIteration:
            raise ValueError(f'Element with symbol "{symbol}" not found')
        return element

    @classmethod
    def from_atomic_number(cls, number: int) -> Type['Element']:
        """
        get Element class by its number
        """
        try:
            element = next(x for x in Element.__subclasses__() if x.atomic_number.fget(None) == number)
        except StopIteration:
            raise ValueError(f'Element with number "{number}" not found')
        return element

    def __eq__(self, other):
        """
        compare attached to molecules elements
        """
        return isinstance(other, Element) and self.atomic_number == other.atomic_number and \
            self.isotope == other.isotope and self.charge == other.charge and self.is_radical == other.is_radical

    def __hash__(self):
        """
        21bit = 9bit | 7bit | 4bit | 1bit
        """
        return (self.isotope or 0) << 12 | self.atomic_number << 5 | self.charge + 4 << 1 | self.is_radical

    def __setstate__(self, state):
        if 'charge' in state:  # 3.1
            self._backward = (state['charge'], state['multiplicity'])
            if isotopes[self.atomic_number - 1] == state['isotope']:
                state['isotope'] = None
        super().__setstate__(state)

    def valence_rules(self, charge: int, is_radical: bool, valence: int) -> \
            List[Tuple[Set[Tuple[int, 'Element']], Dict[Tuple[int, 'Element'], int], int]]:
        """
        valence rules for element with specific charge/radical state
        """
        try:
            return self._compiled_valence_rules[(charge, is_radical, valence)]
        except KeyError:
            raise ValenceError

    @property
    @abstractmethod
    def _common_valences(self) -> Tuple[int, ...]:
        """
        common valences of element
        """

    @property
    @abstractmethod
    def _valences_exceptions(self) -> Tuple[Tuple[int, bool, int, Tuple[Tuple[int, str], ...]]]:
        """
        exceptions in charges, radical state, implicit H count, and non H neighbors of element
        examples:
        (-1, False, 1, ()) - anion, not radical, has 1 implicit hydrogen if explicit atom not exists: [OH]- or C[O-]
        (0, True, 1, ()) - neutral, radical, has 1 implicit hydrogen if explicit atom not exists: [OH]* or C[O*]
        number of free electrons calculated as diff of default valence and number of connected atoms (include implicit)
        (0, False, 1, ((1, 'C'),)) - neutral, not radical, has 1 implicit hydrogen and 1 single bonded carbon:
        CO - alcohol or e.g. COC - ether
        (0, False, 0, ((1, 'O'), (2, 'O'))) - can be anion/cation or neutral: HNO2 or [NO2]-.
        state of neighbors atoms don't take into account. order of neighbors atoms don't take into account.
        use both: (0, False, 0, ((2, 'O'),)) and (0, False, 0, ((1, 'O'), (1, 'O'))) if chains possible
        use charge transfer for carbonyles, cyanides etc:
        (-1, False, 0, ((1, 'C'),)) - [M-]-C#[O+]
        user both for cyanates and isocyanates etc complexes:
        (-1, False, 0, ((1, 'O'),)) and (-1, False, 0, ((1, 'N'),))
        """

    @class_cached_property
    def _compiled_charge_radical(self) -> Set[Tuple[int, bool]]:
        """
        exceptions in charges, radical state
        examples:
        (-1, False) - anion, not radical
        (0, True) - neutral radical
        """
        return {(c, r) for c, r, *_ in self._valences_exceptions}

    @class_cached_property
    def _compiled_valence_rules(self) -> \
            Dict[Tuple[int, bool, int], List[Tuple[Set[Tuple[int, 'Element']], Dict[Tuple[int, 'Element'], int], int]]]:
        """
        dictionary with key = (charge, is_radical, sum_of_bonds) and
        value = list of possible neighbors and implicit H count
        """
        elements_classes = {x.__name__: x.atomic_number.fget(None) for x in Element.__subclasses__()}

        rules = defaultdict(list)
        if self._common_valences[0] and self.atomic_number != 1:  # atom has implicit hydrogens by default except H.
            prev = -1
            for valence in self._common_valences:
                for h in range(valence - prev):
                    rules[(0, False, valence - h)].append((set(), {}, h))  # any atoms and bonds possible
                prev = valence
        else:
            for valence in self._common_valences:
                rules[(0, False, valence)].append((set(), {}, 0))  # any atoms and bonds possible

        for charge, is_radical, implicit, environment in self._valences_exceptions:
            explicit = sum(x for x, _ in environment)
            explicit_dict = defaultdict(int)
            explicit_set = set()
            for b, e in environment:
                be = (b, elements_classes[e])
                explicit_set.add(be)
                explicit_dict[be] += 1
            explicit_dict = dict(explicit_dict)

            if implicit:
                valence = explicit + implicit

                for h in range(implicit + 1):
                    rules[(charge, is_radical, valence - h)].append((explicit_set, explicit_dict, h))
            else:
                rules[(charge, is_radical, explicit)].append((explicit_set, explicit_dict, 0))
        return dict(rules)


# averaged isotopes. 3.*-compatibility
isotopes = tuple(map(int, '''  1                                                                   4
                               7   9                                          11  12  14  16  19  20
                              23  24                                          27  28  31  32  35  40
                              39  40  45  48  51  52  55  56  59  59  64  65  70  73  75  79  80  84
                              85  88  89  91  93  96  98 101 103 106 108 112 115 119 122 128 127 131
                             133 137 139

                                     140 141 144 145 150 152 157 159 163 165 167 169 173 175

                                         178 181 184 186 190 192 195 197 201 204 207 209 209 210 222 
                             223 226 227

                                     232 231 238 237 244 243 247 247 251 252 257 258 259 260

                                         261 270 269 270 270 278 281 281 285 278 289 289 293 297 294
                          '''.split()))


__all__ = ['Element']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import cached_args_method, cached_property
from collections import defaultdict
from typing import List, Union, Tuple, Optional, Dict
from . import cgr, query  # cyclic imports resolve
from .bonds import Bond, DynamicBond
from .common import Graph
from ..algorithms.aromatics import Aromatize
from ..algorithms.calculate2d import Calculate2DMolecule
from ..algorithms.components import StructureComponents
from ..algorithms.depict import DepictMolecule
from ..algorithms.smiles import MoleculeSmiles
from ..algorithms.standardize import Standardize
from ..algorithms.stereo import MoleculeStereo
from ..algorithms.x3dom import X3domMolecule
from ..exceptions import ValenceError, MappingError
from ..periodictable import Element, QueryElement


class MoleculeContainer(MoleculeStereo, Graph, Aromatize, Standardize, MoleculeSmiles, StructureComponents,
                        DepictMolecule, Calculate2DMolecule, X3domMolecule):
    __slots__ = ('_conformers', '_hybridizations', '_atoms_stereo', '_hydrogens', '_cis_trans_stereo',
                 '_allenes_stereo')

    def __init__(self):
        self._conformers: List[Dict[int, Tuple[float, float, float]]] = []
        self._hybridizations: Dict[int, int] = {}
        self._hydrogens: Dict[int, Optional[int]] = {}
        self._atoms_stereo: Dict[int, bool] = {}
        self._allenes_stereo: Dict[int, bool] = {}
        self._cis_trans_stereo: Dict[Tuple[int, int], bool] = {}

        super().__init__()

    def add_atom(self, atom: Union[Element, int, str], *args, charge=0, is_radical=False, **kwargs):
        """
        Add new atom.
        """
        if not isinstance(atom, Element):
            if isinstance(atom, str):
                atom = Element.from_symbol(atom)()
            elif isinstance(atom, int):
                atom = Element.from_atomic_number(atom)()
            else:
                raise TypeError('Element object expected')

        _map = super().add_atom(atom, *args, charge=charge, is_radical=is_radical, **kwargs)
        self._hybridizations[_map] = 1
        self._conformers.clear()  # clean conformers. need full recalculation for new system

        if atom.atomic_number != 1:
            try:
                rules = atom.valence_rules(charge, is_radical, 0)
            except ValenceError:
                self._hydrogens[_map] = None
            else:
                for s, d, h in rules:
                    if h and not s:
                        self._hydrogens[_map] = h
                        break
                else:
                    self._hydrogens[_map] = 0
        else:
            self._hydrogens[_map] = 0
        return _map

    def add_bond(self, n, m, bond: Union[Bond, int]):
        """
        Connect atoms with bonds.

        For Thiele forms of molecule causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.
        """
        if not isinstance(bond, Bond):
            bond = Bond(bond)

        super().add_bond(n, m, bond)
        self._conformers.clear()  # clean conformers. need full recalculation for new system

        self._calc_implicit(n)
        self._calc_implicit(m)

        if bond.order != 1:  # 1 is neutral. 8 is rare. skip.
            self._calc_hybridization(n)
            self._calc_hybridization(m)

        if self._atoms[n].atomic_number != 1 and self._atoms[m].atomic_number != 1:  # not hydrogen
            # fix stereo if formed not to hydrogen bond
            self._fix_stereo()

    def delete_atom(self, n):
        """
        Remove atom.

        For Thiele forms of molecule causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.
        """
        old_bonds = self._bonds[n]  # save bonds
        isnt_hydrogen = self._atoms[n].atomic_number != 1
        super().delete_atom(n)

        del self._hybridizations[n]
        del self._hydrogens[n]
        self._conformers.clear()  # clean conformers. need full recalculation for new system

        for m in old_bonds:
            self._calc_implicit(m)

        for m in old_bonds:
            self._calc_hybridization(m)
        if isnt_hydrogen:  # hydrogen atom not used for stereo coding
            self._fix_stereo()

    def delete_bond(self, n, m):
        """
        Disconnect atoms.

        For Thiele forms of molecule causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.
        """
        super().delete_bond(n, m)
        self._conformers.clear()  # clean conformers. need full recalculation for new system

        self._calc_implicit(n)
        self._calc_implicit(m)
        self._calc_hybridization(n)
        self._calc_hybridization(m)

        if self._atoms[n].atomic_number != 1 and self._atoms[m].atomic_number != 1:
            self._fix_stereo()

    def remap(self, mapping, *, copy=False) -> 'MoleculeContainer':
        h = super().remap(mapping, copy=copy)
        mg = mapping.get
        shg = self._hydrogens

        if copy:
            hh = h._hybridizations
            hhg = h._hydrogens
            hc = h._conformers
            has = h._atoms_stereo
            hal = h._allenes_stereo
            hcs = h._cis_trans_stereo
        else:
            hh = {}
            hhg = {}
            hc = []
            has = {}
            hal = {}
            hcs = {}

        for n, hyb in self._hybridizations.items():
            m = mg(n, n)
            hh[m] = hyb
            hhg[m] = shg[n]

        hc.extend({mg(n, n): x for n, x in c.items()} for c in self._conformers)

        for n, stereo in self._atoms_stereo.items():
            has[mg(n, n)] = stereo
        for n, stereo in self._allenes_stereo.items():
            hal[mg(n, n)] = stereo
        for (n, m), stereo in self._cis_trans_stereo.items():
            hcs[(mg(n, n), mg(m, m))] = stereo

        if copy:
            return h

        self._hybridizations = hh
        self._hydrogens = hhg
        self._conformers = hc
        self._atoms_stereo = has
        self._allenes_stereo = hal
        self._cis_trans_stereo = hcs
        return self

    def copy(self, **kwargs) -> 'MoleculeContainer':
        copy = super().copy(**kwargs)
        copy._hybridizations = self._hybridizations.copy()
        copy._hydrogens = self._hydrogens.copy()
        copy._conformers = [c.copy() for c in self._conformers]
        copy._atoms_stereo = self._atoms_stereo.copy()
        copy._allenes_stereo = self._allenes_stereo.copy()
        copy._cis_trans_stereo = self._cis_trans_stereo.copy()
        return copy

    def substructure(self, atoms, *, as_query: bool = False, **kwargs) -> Union['MoleculeContainer',
                                                                                'query.QueryContainer']:
        """
        Create substructure containing atoms from atoms list.

        For Thiele forms of molecule In Molecule substructure causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.

        :param atoms: list of atoms numbers of substructure
        :param meta: if True metadata will be copied to substructure
        :param as_query: return Query object based on graph substructure
        """
        sub, atoms = super().substructure(atoms, query.QueryContainer if as_query else self.__class__, **kwargs)
        sa = self._atoms
        sb = self._bonds

        if as_query:
            lost = {n for n, a in sa.items() if a.atomic_number != 1} - set(atoms)  # atoms not in substructure
            not_skin = {n for n in atoms if lost.isdisjoint(sb[n])}
            sub._atoms_stereo = {n: s for n, s in self._atoms_stereo.items() if n in not_skin}
            sub._allenes_stereo = {n: s for n, s in self._allenes_stereo.items()
                                   if not_skin.issuperset(self._stereo_allenes_paths[n]) and
                                      not_skin.issuperset(x for x in self._stereo_allenes[n] if x)}
            sub._cis_trans_stereo = {nm: s for nm, s in self._cis_trans_stereo.items()
                                     if not_skin.issuperset(self._stereo_cis_trans_paths[nm]) and
                                        not_skin.issuperset(x for x in self._stereo_cis_trans[nm] if x)}
            sub._atoms = ca = {}
            for n in atoms:
                atom = sa[n]
                atom = QueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
                ca[n] = atom
                atom._attach_to_graph(sub, n)

            sh = self._hybridizations
            sub._neighbors = {n: (len(sb[n]),) for n in atoms}
            sub._hybridizations = {n: (sh[n],) for n in atoms}
        else:
            sub._conformers = [{n: c[n] for n in atoms} for c in self._conformers]
            sub._atoms = ca = {}
            for n in atoms:
                atom = sa[n].copy()
                ca[n] = atom
                atom._attach_to_graph(sub, n)

            # recalculate query marks
            sub._hybridizations = {}
            sub._hydrogens = {}
            for n in sub._atoms:
                sub._calc_hybridization(n)
                sub._calc_implicit(n)
            # fix_stereo will repair data
            sub._atoms_stereo = self._atoms_stereo
            sub._allenes_stereo = self._allenes_stereo
            sub._cis_trans_stereo = self._cis_trans_stereo
            sub._fix_stereo()
        return sub

    def union(self, other, **kwargs):
        if isinstance(other, MoleculeContainer):
            u, other = super().union(other, **kwargs)
            u._conformers.clear()

            u._hybridizations.update(other._hybridizations)
            u._hydrogens.update(other._hydrogens)
            u._atoms_stereo.update(other._atoms_stereo)
            u._allenes_stereo.update(other._allenes_stereo)
            u._cis_trans_stereo.update(other._cis_trans_stereo)

            ub = u._bonds
            for n, m_bond in other._bonds.items():
                ub[n] = ubn = {}
                for m, bond in m_bond.items():
                    if m in ub:  # bond partially exists. need back-connection.
                        ubn[m] = ub[m][n]
                    else:
                        ubn[m] = bond.copy()

            ua = u._atoms
            for n, atom in other._atoms.items():
                atom = atom.copy()
                ua[n] = atom
                atom._attach_to_graph(u, n)
            return u
        elif isinstance(other, Graph):
            return other.union(self, **kwargs)
        else:
            raise TypeError('Graph expected')

    def compose(self, other: Union['MoleculeContainer', 'cgr.CGRContainer']) -> 'cgr.CGRContainer':
        """
        Compose 2 graphs to CGR.
        """
        sa = self._atoms
        sc = self._charges
        sr = self._radicals
        sp = self._plane
        sb = self._bonds

        bonds = []
        adj = defaultdict(lambda: defaultdict(lambda: [None, None]))

        if isinstance(other, MoleculeContainer):
            oa = other._atoms
            oc = other._charges
            or_ = other._radicals
            op = other._plane
            ob = other._bonds
            common = sa.keys() & other
            h = cgr.CGRContainer()
            atoms = h._atoms

            for n in sa.keys() - common:  # cleavage atoms
                h.add_atom(sa[n], n, charge=sc[n], is_radical=sr[n], xy=sp[n], p_charge=sc[n], p_is_radical=sr[n])
                for m, bond in sb[n].items():
                    if m not in atoms:
                        if m in common:  # bond to common atoms is broken bond
                            order = bond.order
                            bond = object.__new__(DynamicBond)
                            bond._DynamicBond__order, bond._DynamicBond__p_order = order, None
                        bonds.append((n, m, bond))
            for n in other._atoms.keys() - common:  # coupling atoms
                h.add_atom(oa[n], n, charge=oc[n], is_radical=or_[n], xy=op[n], p_charge=oc[n], p_is_radical=or_[n])
                for m, bond in ob[n].items():
                    if m not in atoms:
                        if m in common:  # bond to common atoms is formed bond
                            order = bond.order
                            bond = object.__new__(DynamicBond)
                            bond._DynamicBond__order, bond._DynamicBond__p_order = None, order
                        bonds.append((n, m, bond))
            for n in common:
                an = adj[n]
                for m, bond in sb[n].items():
                    if m in common:
                        an[m][0] = bond.order
                for m, bond in ob[n].items():
                    if m in common:
                        an[m][1] = bond.order
            for n in common:
                san = sa[n]
                if san.atomic_number != oa[n].atomic_number or san.isotope != oa[n].isotope:
                    raise MappingError(f'atoms with number {{{n}}} not equal')
                h.add_atom(san, n, charge=sc[n], is_radical=sr[n], xy=sp[n], p_charge=oc[n], p_is_radical=or_[n])
                for m, (o1, o2) in adj[n].items():
                    if m not in atoms:
                        bond = object.__new__(DynamicBond)
                        bond._DynamicBond__order, bond._DynamicBond__p_order = o1, o2
                        bonds.append((n, m, bond))
        elif isinstance(other, cgr.CGRContainer):
            oa = other._atoms
            oc = other._charges
            or_ = other._radicals
            opc = other._p_charges
            opr = other._p_radicals
            op = other._plane
            ob = other._bonds
            common = sa.keys() & other
            h = other.__class__()  # subclasses support
            atoms = h._atoms

            for n in sa.keys() - common:  # cleavage atoms
                h.add_atom(sa[n], n, charge=sc[n], is_radical=sr[n], xy=sp[n], p_charge=sc[n], p_is_radical=sr[n])
                for m, bond in sb[n].items():
                    if m not in atoms:
                        if m in common:  # bond to common atoms is broken bond
                            order = bond.order
                            bond = object.__new__(DynamicBond)
                            bond._DynamicBond__order, bond._DynamicBond__p_order = order, None
                        bonds.append((n, m, bond))
            for n in other._atoms.keys() - common:  # coupling atoms
                h.add_atom(oa[n].copy(), n, charge=oc[n], is_radical=or_[n], xy=op[n], p_charge=opc[n],
                           p_is_radical=opr[n])
                for m, bond in ob[n].items():
                    if m not in atoms:
                        if m in common:  # bond to common atoms is formed bond
                            order = bond.p_order
                            if order:  # skip broken bond. X>None => None>None
                                bond = object.__new__(DynamicBond)
                                bond._DynamicBond__order, bond._DynamicBond__p_order = None, order
                                bonds.append((n, m, bond))
                        else:
                            bonds.append((n, m, bond))
            for n in common:
                an = adj[n]
                for m, bond in sb[n].items():
                    if m in common:
                        an[m][0] = bond.order
                for m, bond in ob[n].items():
                    if m in an or m in common and bond.p_order:
                        # self has nm bond or other bond not broken
                        an[m][1] = bond.p_order
            for n in common:
                san = sa[n]
                if san.atomic_number != oa[n].atomic_number or san.isotope != oa[n].isotope:
                    raise MappingError(f'atoms with number {{{n}}} not equal')
                h.add_atom(san, n, charge=sc[n], is_radical=sr[n], xy=sp[n], p_charge=opc[n], p_is_radical=opr[n])
                for m, (o1, o2) in adj[n].items():
                    if m not in atoms:
                        bond = object.__new__(DynamicBond)
                        bond._DynamicBond__order, bond._DynamicBond__p_order = o1, o2
                        bonds.append((n, m, bond))
        else:
            raise TypeError('MoleculeContainer or CGRContainer expected')

        for n, m, bond in bonds:
            h.add_bond(n, m, bond)
        return h

    def __xor__(self, other):
        """
        G ^ H is CGR generation
        """
        return self.compose(other)

    def get_mapping(self, other: 'MoleculeContainer', **kwargs):
        if isinstance(other, MoleculeContainer):
            return super().get_mapping(other, **kwargs)
        raise TypeError('MoleculeContainer expected')

    def get_mcs_mapping(self, other: 'MoleculeContainer', **kwargs):
        if isinstance(other, MoleculeContainer):
            return super().get_mcs_mapping(other, **kwargs)
        raise TypeError('MoleculeContainer expected')

    @cached_property
    def molecular_charge(self):
        """
        Total charge of molecule
        """
        return sum(self._charges.values())

    def __int__(self):
        """
        Total charge of molecule
        """
        return self.molecular_charge

    @cached_property
    def molecular_mass(self):
        return sum(x.atomic_mass for x in self._atoms.values())

    def __float__(self):
        return self.molecular_mass

    @cached_args_method
    def _explicit_hydrogens(self, n: int) -> int:
        """
        Number of explicit hydrogen atoms connected to atom.

        Take into account any type of bonds with hydrogen atoms.
        """
        atoms = self._atoms
        return sum(atoms[m].atomic_number == 1 for m in self._bonds[n])

    @cached_args_method
    def _total_hydrogens(self, n: int) -> int:
        return self._hydrogens[n] + self._explicit_hydrogens(n)

    def _calc_implicit(self, n: int):
        atoms = self._atoms
        atom = atoms[n]
        if atom.atomic_number != 1:
            charge = self._charges[n]
            is_radical = self._radicals[n]
            explicit_sum = 0
            explicit_dict = defaultdict(int)
            for m, bond in self._bonds[n].items():
                order = bond.order
                if order == 4:  # aromatic rings not supported
                    self._hydrogens[n] = None
                    return
                elif order != 8:  # any bond used for complexes
                    explicit_sum += order
                    explicit_dict[(order, atoms[m].atomic_number)] += 1
            try:
                rules = atom.valence_rules(charge, is_radical, explicit_sum)
            except ValenceError:
                self._hydrogens[n] = None
                return
            for s, d, h in rules:
                if h and s.issubset(explicit_dict) and all(explicit_dict[k] >= c for k, c in d.items()):
                    self._hydrogens[n] = h
                    return
        self._hydrogens[n] = 0

    def _calc_hybridization(self, n: int):
        hybridization = 1
        for bond in self._bonds[n].values():
            order = bond.order
            if order == 4:
                self._hybridizations[n] = 4
                return
            elif order == 3:
                if hybridization != 3:
                    hybridization = 3
            elif order == 2:
                if hybridization == 2:
                    hybridization = 3
                elif hybridization == 1:
                    hybridization = 2
        self._hybridizations[n] = hybridization

    def __getstate__(self):
        return {'conformers': self._conformers, 'atoms_stereo': self._atoms_stereo,
                'allenes_stereo': self._allenes_stereo, 'cis_trans_stereo': self._cis_trans_stereo,
                **super().__getstate__()}

    def __setstate__(self, state):
        if '_BaseContainer__meta' in state:  # 2.8 reverse compatibility
            state['atoms'] = {n: Element.from_symbol(a['element'])(a.get('isotope')) for n, a in state['_node'].items()}
            state['charges'] = {n: a['s_charge'] for n, a in state['_node'].items()}
            state['radicals'] = {n: bool(a['s_radical']) for n, a in state['_node'].items()}
            state['plane'] = {n: (a['s_x'], a['s_y']) for n, a in state['node'].items()}

            state['bonds'] = bonds = {}
            for n, m_bond in state['_adj'].items():
                bonds[n] = bn = {}
                for m, bond in m_bond.items():
                    if m in bonds:
                        bn[m] = bonds[m][n]
                    else:
                        bn[m] = Bond(bond['s_bond'] if bond['s_bond'] != 9 else 5)

            state['conformers'] = []
            state['atoms_stereo'] = {}
            state['allenes_stereo'] = {}
            state['cis_trans_stereo'] = {}
            state['meta'] = state['_BaseContainer__meta']
            state['parsed_mapping'] = {}
        elif 'node' in state:  # 3.1 compatibility.
            state['atoms'] = {n: a.atom.copy() for n, a in state['node'].items()}
            state['charges'] = {n: a.charge for n, a in state['node'].items()}
            state['radicals'] = {n: a.is_radical for n, a in state['node'].items()}
            state['plane'] = {n: a.xy for n, a in state['node'].items()}
            state['parsed_mapping'] = {n: a.parsed_mapping for n, a in state['node'].items() if a.parsed_mapping}

            state['bonds'] = bonds = {}
            for n, m_bond in state['adj'].items():
                bonds[n] = bn = {}
                for m, bond in m_bond.items():
                    if m in bonds:
                        bn[m] = bonds[m][n]
                    else:
                        bn[m] = Bond(bond.order)

            state['conformers'] = []
            state['atoms_stereo'] = {}
            state['allenes_stereo'] = {}
            state['cis_trans_stereo'] = {}
        elif 'allenes_stereo' not in state:  # <4.0.22
            state['atoms_stereo'] = {}  # flush existing stereo if exists.
            state['allenes_stereo'] = {}
            state['cis_trans_stereo'] = {}

        super().__setstate__(state)
        self._conformers = state['conformers']
        self._atoms_stereo = state['atoms_stereo']
        self._allenes_stereo = state['allenes_stereo']
        self._cis_trans_stereo = state['cis_trans_stereo']

        # restore query and hydrogen marks
        self._hybridizations = {}
        self._hydrogens = {}
        for n in state['bonds']:
            self._calc_hybridization(n)
            self._calc_implicit(n)


__all__ = ['MoleculeContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import Tuple, Dict, Type
from .core import Core
from .dynamic import Dynamic, DynamicElement
from ...exceptions import IsNotConnectedAtom


class DynamicQuery(Dynamic):
    __slots__ = ()

    @property
    def neighbors(self) -> Tuple[int, ...]:
        """
        Number of neighbors of atom in reactant state.
        """
        try:
            return self._graph()._neighbors[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def p_neighbors(self) -> Tuple[int, ...]:
        """
        Number of neighbors of atom in product state.
        """
        try:
            return self._graph()._p_neighbors[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @neighbors.setter
    def neighbors(self, neighbors):
        try:
            g = self._graph()
            neighbors = g._validate_neighbors(neighbors)
            neighbors, p_neighbors = g._validate_neighbors_pairing(neighbors, g._p_neighbors[self._map])
            g._neighbors[self._map] = neighbors
            g._p_neighbors[self._map] = p_neighbors
            g.flush_cache()
        except AttributeError:
            raise IsNotConnectedAtom

    @p_neighbors.setter
    def p_neighbors(self, p_neighbors):
        try:
            g = self._graph()
            p_neighbors = g._validate_neighbors(p_neighbors)
            neighbors, p_neighbors = g._validate_neighbors_pairing(g._neighbors[self._map], p_neighbors)
            g._neighbors[self._map] = neighbors
            g._p_neighbors[self._map] = p_neighbors
            g.flush_cache()
        except AttributeError:
            raise IsNotConnectedAtom

    @Core.hybridization.setter
    def hybridization(self, hybridization):
        try:
            g = self._graph()
            hybridization = g._validate_hybridization(hybridization)
            hybridization, p_hybridization = g._validate_hybridization_pairing(hybridization,
                                                                               g._p_hybridizations[self._map])
            g._hybridizations[self._map] = hybridization
            g._p_hybridizations[self._map] = p_hybridization
            g.flush_cache()
        except AttributeError:
            raise IsNotConnectedAtom

    @Dynamic.p_hybridization.setter
    def p_hybridization(self, p_hybridization):
        try:
            g = self._graph()
            p_hybridization = g._validate_hybridization(p_hybridization)
            hybridization, p_hybridization = g._validate_hybridization_pairing(g._hybridizations[self._map],
                                                                               p_hybridization)
            g._hybridizations[self._map] = hybridization
            g._p_hybridizations[self._map] = p_hybridization
            g.flush_cache()
        except AttributeError:
            raise IsNotConnectedAtom

    _hybridization_bitmap = {(): 0, ((1, 1),): 1, ((2, 2),): 2, ((3, 3),): 3, ((4, 4),): 4,
                             ((1, 2),): 5, ((1, 3),): 6, ((1, 4),): 7, ((2, 3),): 8, ((2, 4),): 9, ((3, 4),): 10,
                             ((2, 1),): 11, ((3, 1),): 12, ((4, 1),): 13, ((3, 2),): 14,
                             ((4, 2),): 15, ((4, 3),): 16, }  # 31 is any other combination
    _neighbors_bitmap = {(): 0, ((1, 1),): 1, ((2, 2),): 2, ((3, 3),): 3, ((4, 4),): 4, ((5, 5),): 5, ((6, 6),): 6,
                         ((1, 2),): 7, ((2, 3),): 8, ((3, 4),): 9, ((4, 5),): 10, ((5, 6),): 11, ((2, 1),): 12,
                         ((3, 2),): 13, ((4, 3),): 14, ((5, 4),): 15, ((6, 5),): 16}  # 31 is any other combination


class DynamicQueryElement(DynamicQuery):
    __slots__ = ()

    @property
    def atomic_symbol(self) -> str:
        return self.__class__.__name__[12:]

    @classmethod
    def from_symbol(cls, symbol: str) -> Type['DynamicQueryElement']:
        """
        get Element class by its symbol
        """
        if symbol == 'A':
            return DynamicAnyElement
        try:
            element = next(x for x in DynamicQueryElement.__subclasses__() if x.__name__ == f'DynamicQuery{symbol}')
        except StopIteration:
            raise ValueError(f'DynamicQueryElement with symbol "{symbol}" not found')
        return element

    @classmethod
    def from_atomic_number(cls, number: int) -> Type['DynamicQueryElement']:
        """
        get Element class by its number
        """
        if number == 0:
            return DynamicAnyElement
        try:
            element = next(x for x in DynamicQueryElement.__subclasses__() if x.atomic_number.fget(None) == number)
        except StopIteration:
            raise ValueError(f'DynamicQueryElement with number "{number}" not found')
        return element

    def __eq__(self, other):
        if isinstance(other, DynamicElement):
            if self.atomic_number == other.atomic_number and \
                    self.charge == other.charge and self.p_charge == other.p_charge and \
                    self.is_radical == other.is_radical and self.p_is_radical == other.p_is_radical:
                if self.isotope and self.isotope != other.isotope:
                    return False
                if self.neighbors and (other.neighbors, other.p_neighbors) not in zip(self.neighbors, self.p_neighbors):
                    return False
                if self.hybridization and (other.hybridization, other.p_hybridization) not in zip(self.hybridization,
                                                                                                  self.p_hybridization):
                    return False
                return True
        elif isinstance(other, DynamicQueryElement) and self.atomic_number == other.atomic_number and \
                self.isotope == other.isotope and self.charge == other.charge and self.p_charge == other.p_charge and \
                self.is_radical == other.is_radical and self.p_is_radical == other.p_is_radical and \
                self.neighbors == other.neighbors and self.hybridization and other.hybridization and \
                self.p_neighbors == other.p_neighbors and self.p_hybridization and other.p_hybridization:
            # equal query element has equal query marks
            return True
        return False

    def __hash__(self):
        """
        36bit = 9bit | 7bit | 4bit | 4bit | 1bit | 1bit | 5bit | 5bit
        """
        return (self.isotope or 0) << 27 | self.atomic_number << 20 | self.charge + 4 << 16 | self.p_charge << 12 | \
            self.is_radical << 11 | self.p_is_radical << 10 | \
            self._hybridization_bitmap.get(tuple(zip(self.hybridization, self.p_hybridization)), 31) << 5 | \
            self._neighbors_bitmap.get(tuple(zip(self.neighbors, self.p_neighbors)), 31)


class DynamicAnyElement(DynamicQuery):  # except Hydrogen!
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__()

    @property
    def atomic_symbol(self) -> str:
        return 'A'

    @property
    def atomic_number(self) -> int:
        return 0

    @property
    def isotopes_distribution(self) -> Dict[int, float]:
        return {}

    @property
    def isotopes_masses(self) -> Dict[int, float]:
        return {}

    @property
    def atomic_radius(self):
        return 0.5

    def __eq__(self, other):
        if isinstance(other, DynamicElement):
            if self.charge == other.charge and self.p_charge == other.p_charge and \
                    self.is_radical == other.is_radical and self.p_is_radical == other.p_is_radical:
                if self.neighbors and (other.neighbors, other.p_neighbors) not in zip(self.neighbors, self.p_neighbors):
                    return False
                if self.hybridization and (other.hybridization, other.p_hybridization) not in zip(self.hybridization,
                                                                                                  self.p_hybridization):
                    return False
                return True
        elif isinstance(other, DynamicQuery) and self.charge == other.charge and self.p_charge == other.p_charge and \
                self.is_radical == other.is_radical and self.p_is_radical == other.p_is_radical and \
                self.neighbors == other.neighbors and self.hybridization and other.hybridization and \
                self.p_neighbors == other.p_neighbors and self.p_hybridization and other.p_hybridization:
            return True
            # equal query element has equal query marks
        return False

    def __hash__(self):
        """
        20bit = 4bit | 4bit | 1bit | 1bit | 5bit | 5bit
        """
        return self.charge + 4 << 16 | self.p_charge << 12 | \
            self.is_radical << 11 | self.p_is_radical << 10 | \
            self._hybridization_bitmap.get(tuple(zip(self.hybridization, self.p_hybridization)), 31) << 5 | \
            self._neighbors_bitmap.get(tuple(zip(self.neighbors, self.p_neighbors)), 31)


__all__ = ['DynamicQueryElement', 'DynamicAnyElement']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import cached_property
from collections import defaultdict
from itertools import chain, combinations, product
from logging import warning
from operator import itemgetter
from typing import Any, Dict, Set, Tuple, Union
from ..exceptions import ImplementationError


class SSSR:
    """ SSSR calculation. based on idea of PID matrices from:
        Lee, C. J., Kang, Y.-M., Cho, K.-H., & No, K. T. (2009).
        A robust method for searching the smallest set of smallest rings with a path-included distance matrix.
        Proceedings of the National Academy of Sciences of the United States of America, 106(41), 17355–17358.
        http://doi.org/10.1073/pnas.0813040106
    """
    __slots__ = ()

    @cached_property
    def sssr(self) -> Tuple[Tuple[int, ...], ...]:
        """
        Smallest Set of Smallest Rings.

        :return rings atoms numbers
        """
        if self.rings_count:
            return self._sssr(self._bonds, self.rings_count)
        return ()

    @classmethod
    def _sssr(cls, bonds: Dict[int, Union[Set[int], Dict[int, Any]]], n_sssr: int) -> Tuple[Tuple[int, ...], ...]:
        """
        Smallest Set of Smallest Rings of any adjacency matrix.
        Number of rings required.
        """
        bonds = cls._skin_graph(bonds)
        return cls.__rings_filter(cls.__c_set(*cls.__make_pid(bonds)), n_sssr, bonds)

    @staticmethod
    def __make_pid(bonds):
        lb = len(bonds)
        pid1 = defaultdict(lambda: defaultdict(dict))
        pid2 = defaultdict(lambda: defaultdict(dict))
        distances = defaultdict(lambda: defaultdict(lambda: lb))
        for n, ms in bonds.items():
            dn = distances[n]
            pn = pid1[n]
            for m in ms:
                pn[m][(m, n)] = (n, m)
                dn[m] = 1

        for k in bonds:
            new_distances = defaultdict(dict)
            dk = distances[k]
            ndk = new_distances[k]
            for i in bonds:
                if i == k:
                    continue
                di = distances[i]
                ndi = new_distances[i]
                ndk[i] = ndi[k] = di[k]
                for j in bonds:
                    if j == k or j == i:
                        continue
                    ij = di[j]
                    ikj = di[k] + dk[j]
                    if ij - ikj == 1:  # A new shortest path == previous shortest path - 1
                        pid2[i][j] = pid1[i][j]
                        pid1[i][j] = {(ni, mj): ip[:-1] + jp for ((ni, _), ip), ((_, mj), jp) in
                                      product(pid1[i][k].items(), pid1[k][j].items())}
                        ndi[j] = ikj
                    elif ij > ikj:  # A new shortest path
                        pid2[i][j] = {}
                        pid1[i][j] = {(ni, mj): ip[:-1] + jp for ((ni, _), ip), ((_, mj), jp) in
                                      product(pid1[i][k].items(), pid1[k][j].items())}
                        ndi[j] = ikj
                    elif ij == ikj:  # Another shortest path
                        pid1[i][j].update({(ni, mj): ip[:-1] + jp for ((ni, _), ip), ((_, mj), jp) in
                                           product(pid1[i][k].items(), pid1[k][j].items())})
                        ndi[j] = ij
                    elif ikj - ij == 1:  # Shortest+1 path
                        pid2[i][j].update({(ni, mj): ip[:-1] + jp for ((ni, _), ip), ((_, mj), jp) in
                                           product(pid1[i][k].items(), pid1[k][j].items())})
                        ndi[j] = ij
                    else:
                        ndi[j] = ij
            distances = new_distances
        return pid1, pid2, distances

    @staticmethod
    def __c_set(pid1, pid2, pid1l):
        c_set = []
        seen = set()
        for i, p1i in pid1.items():
            seen.add(i)
            di = pid1l[i]
            p2i = pid2[i]

            for j, p1ij in p1i.items():
                if j in seen:
                    continue
                p1ij = list(p1ij.values())
                p2ij = list(p2i[j].values())
                dij = di[j] * 2

                if len(p1ij) == 1:  # one shortest
                    if not p2ij:  # need shortest + 1 path
                        continue
                    c_set.append((dij + 1, p1ij, p2ij))
                elif not p2ij:  # one or more odd rings
                    c_set.append((dij, p1ij, None))
                else:  # odd and even rings found (e.g. bicycle)
                    c_set.append((dij, p1ij, None))
                    c_set.append((dij + 1, p1ij, p2ij))

        for c_num, p1ij, p2ij in sorted(c_set, key=itemgetter(0)):
            if c_num % 2:  # odd rings
                for c1 in p1ij:
                    for c2 in p2ij:
                        yield c1 + c2[-2:0:-1]
            else:
                for c1, c2 in zip(p1ij, p1ij[1:]):
                    yield c1 + c2[-2:0:-1]

    @staticmethod
    def __rings_filter(rings, n_sssr, bonds):
        # step 1: collect isolated rings
        c = next(rings)
        if n_sssr == 1:
            return c,

        ck = frozenset(c)
        seen_rings = {ck}
        sssr = {ck: c}
        condensed_rings = {ck}  # collection of contours of condensed rings
        hold = []
        for c in rings:
            ck = frozenset(c)
            if len(ck) != len(c) or ck in seen_rings:
                continue
            seen_rings.add(ck)

            # create graph of connected neighbour rings
            neighbors = {x: set() for x in chain(sssr, condensed_rings) if len(x & ck) > 1}
            if neighbors:
                for i, j in combinations(neighbors, 2):
                    if len(i & j) > 1:
                        neighbors[i].add(j)
                        neighbors[j].add(i)
                # check if hold rings is combination of existing. (123654) is combo of (1254) and (2365)
                #
                # 1--2--3
                # |  |  |
                # 4--5--6
                #
                # modified NX.dfs_labeled_edges
                # https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.traver\
                # sal.depth_first_search.dfs_labeled_edges.html
                depth_limit = len(neighbors) - 1
                for start, nbrs in neighbors.items():
                    if not nbrs:
                        continue
                    stack = [(start, depth_limit, iter(neighbors[start]), {start})]
                    while stack:
                        parent, depth_now, children, seen = stack[-1]
                        try:
                            child = next(children)
                        except StopIteration:
                            stack.pop()
                        else:
                            if child not in seen:
                                common = parent & child
                                if len(common) > 2:  # only terminal common atoms required
                                    mc = parent ^ child | {n for n in common if len(common.intersection(bonds[n])) == 1}
                                else:  # neighbors at least have 2 common atoms
                                    mc = parent | child
                                if ck == mc:  # macrocycle found
                                    break
                                elif depth_now and 2 < len(mc) <= len(c):
                                    stack.append((mc, depth_now - 1, iter(neighbors[child]), {child} | seen))
                    else:
                        continue
                    break
                else:
                    # update condensed rings. need for cuban and fullerene like structures.
                    tmp = set()
                    ckc = ck
                    for r in condensed_rings:
                        common = r & ckc
                        lc = len(common)
                        if lc == 2:
                            ckc |= r
                        elif lc == len(ckc):
                            unique = r - ckc
                            ckc = unique | {n for n in common if not unique.isdisjoint(bonds[n])}
                        # elif lc == len(r):  # impossible?
                        #     ckc -= r
                        elif lc > 2:
                            ckc = r ^ ckc | {n for n in common if len(common.intersection(bonds[n])) == 1}
                            while True:
                                term = {n for n in ckc if len(ckc.intersection(bonds[n])) == 1}
                                if not term:
                                    break
                                ckc -= term
                        else:
                            tmp.add(r)
                    condensed_rings = tmp

                    if ckc != ck and ckc in seen_rings:
                        # check ring for full surrounding by other rings
                        # reduced to existing ring. finish reached?
                        neighbors = set()  # bonds of neighbors
                        for r in sssr:
                            if len(r & ck) > 1:
                                for n in r:
                                    for m in bonds[n]:
                                        if m in r:
                                            neighbors.add((n, m))
                        if (c[0], c[-1]) in neighbors and all(x in neighbors for x in zip(c, c[1:])):
                            condensed_rings.add(ck)  # add ring to condensed. required for combined rings detection.
                            hold.append(c)
                            continue

                    condensed_rings.add(ckc)
                    seen_rings.add(ckc)
                    sssr[ck] = c
                    if len(sssr) == n_sssr:
                        return tuple(sssr.values())
            else:
                condensed_rings.add(ck)
                sssr[ck] = c
                if len(sssr) == n_sssr:
                    return tuple(sssr.values())

        if len(hold) < n_sssr - len(sssr):
            raise ImplementationError
        elif len(hold) > n_sssr - len(sssr):
            warning('Number of closing rings more than SSSR count. Possible errors.')
        return (*sssr.values(), *hold[:n_sssr - len(sssr)])


__all__ = ['SSSR']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import abstractmethod
from CachedMethods import cached_property
from collections import defaultdict
from itertools import permutations
from typing import Dict, Iterator, Any
from .._functions import lazy_product


frequency = {1: 10,  # H
             6: 9,  # C
             8: 8,  # O
             7: 7,  # N
             15: 6, 16: 6,  # P, S
             9: 5,  # F
             17: 4, 35: 4,  # Cl, Br
             53: 3,  # I
             5: 2, 14: 2,  # B, Si
             11: 1, 12: 1, 19: 1, 20: 1}  # Na, Mag,  K, Ca


def atom_frequency(x):
    return frequency.get(x.atomic_number, 0)


class Isomorphism:
    __slots__ = ()

    def __lt__(self, other):
        if len(self) >= len(other):
            return False
        return self.is_substructure(other)

    def __le__(self, other):
        return self.is_substructure(other)

    def __gt__(self, other):
        if len(self) <= len(other):
            return False
        return other.is_substructure(self)

    def __ge__(self, other):
        return other.is_substructure(self)

    def is_substructure(self, other) -> bool:
        """
        Test self is substructure of other
        """
        try:
            next(self.get_mapping(other))
        except StopIteration:
            return False
        return True

    def is_equal(self, other) -> bool:
        """
        Test self is same structure as other
        """
        if len(self) != len(other):
            return False
        try:
            next(self.get_mapping(other))
        except StopIteration:
            return False
        return True

    @abstractmethod
    def get_mapping(self, other, *, automorphism_filter: bool = True,
                    optimize: bool = True, fallback: bool = False) -> Iterator[Dict[int, int]]:
        """
        Get self to other substructure mapping generator.

        :param automorphism_filter: Skip matches to same atoms.
        :param optimize: Morgan weights based automorphism preventing.
        :param fallback: Try without optimization then nothing matched.
        """
        if optimize:
            g = self.__components_mapping(other, other.atoms_order, automorphism_filter)
            m = next(g, None)
            if m is not None:
                yield m
                yield from g
                return
            elif not fallback:
                return
        yield from self.__components_mapping(other, {n: i for i, n in enumerate(other)}, automorphism_filter)

    def __components_mapping(self, other, o_order, automorphism_filter):
        components, closures = self.__compiled_query
        o_atoms = other._atoms
        o_bonds = other._bonds

        seen = set()
        if len(components) == 1:
            for candidate in other.connected_components:
                for mapping in self.__get_mapping(components[0], closures, o_atoms, o_bonds, set(candidate), o_order):
                    if automorphism_filter:
                        atoms = frozenset(mapping.values())
                        if atoms in seen:
                            continue
                        seen.add(atoms)
                    yield mapping
        else:
            for candidates in permutations((set(x) for x in other.connected_components), len(components)):
                mappers = [self.__get_mapping(order, closures, o_atoms, o_bonds, component, o_order)
                           for order, component in zip(components, candidates)]
                for match in lazy_product(*mappers):
                    mapping = match[0]
                    for m in match[1:]:
                        mapping.update(m)
                    if automorphism_filter:
                        atoms = frozenset(mapping.values())
                        if atoms in seen:
                            continue
                        seen.add(atoms)
                    yield mapping

    @staticmethod
    def __get_mapping(linear_query, query_closures, o_atoms, o_bonds, scope, groups):
        size = len(linear_query) - 1
        order_depth = {v[0]: k for k, v in enumerate(linear_query)}
        equal_cache = defaultdict(dict)

        stack = []
        path = []
        mapping = {}
        reversed_mapping = {}

        s_n, s_atom = linear_query[0]
        eqs = equal_cache[s_n]
        for n, o_atom in o_atoms.items():
            if n in scope:
                if s_atom == o_atom:
                    eqs[n] = True
                    stack.append((n, 0))
                else:
                    eqs[n] = False

        while stack:
            n, depth = stack.pop()
            current = linear_query[depth][0]
            if depth == size:
                yield {current: n, **mapping}
            else:
                if len(path) != depth:
                    for x in path[depth:]:
                        del mapping[reversed_mapping.pop(x)]
                    path = path[:depth]

                path.append(n)
                mapping[current] = n
                reversed_mapping[n] = current

                depth += 1
                s_n, back, s_atom, s_bond = linear_query[depth]
                if back != current:
                    n = path[order_depth[back]]

                eqs = equal_cache[s_n]
                uniq = set()
                for o_n, o_bond in o_bonds[n].items():
                    if o_n in scope and o_n not in reversed_mapping and s_bond == o_bond and groups[o_n] not in uniq:
                        uniq.add(groups[o_n])
                        if o_n in eqs:
                            if eqs[o_n]:
                                if all(bond == o_bonds[mapping[m]].get(o_n) for m, bond in query_closures[s_n]):
                                    stack.append((o_n, depth))
                        elif s_atom == o_atoms[o_n]:
                            eqs[o_n] = True
                            if all(bond == o_bonds[mapping[m]].get(o_n) for m, bond in query_closures[s_n]):
                                stack.append((o_n, depth))
                        else:
                            eqs[o_n] = False

    @cached_property
    def __compiled_query(self):
        return self.__compile_query(self._atoms, self._bonds, {n: atom_frequency(a) for n, a in self._atoms.items()})

    @staticmethod
    def __compile_query(atoms, bonds, atoms_frequencies):
        closures = defaultdict(list)
        components = []
        seen = set()
        while len(seen) < len(atoms):
            start = min(atoms.keys() - seen, key=lambda x: atoms_frequencies[x])
            seen.add(start)
            stack = [(n, start, atoms[n], bond) for n, bond in sorted(bonds[start].items(), reverse=True,
                                                                      key=lambda x: atoms_frequencies[x[0]])]
            order = [(start, atoms[start])]
            components.append(order)

            while stack:
                front, back, *_ = atom = stack.pop()
                if front not in seen:
                    order.append(atom)
                    for n, bond in sorted(bonds[front].items(), reverse=True, key=lambda x: atoms_frequencies[x[0]]):
                        if n != back:
                            if n not in seen:
                                stack.append((n, front, atoms[n], bond))
                            else:
                                closures[front].append((n, bond))
                    seen.add(front)
        return components, closures

    def is_automorphic(self):
        """
        Test for automorphism symmetry of graph.
        """
        try:
            next(self.get_automorphism_mapping())
        except StopIteration:
            return False
        return True

    def get_automorphism_mapping(self) -> Iterator[Dict[int, int]]:
        """
        Iterator of all possible automorphism mappings.
        """
        return self._get_automorphism_mapping(self.atoms_order, self._bonds,
                                              {n: atom_frequency(a) for n, a in self._atoms.items()})

    @classmethod
    def _get_automorphism_mapping(cls, atoms: Dict[int, int], bonds: Dict[int, Dict[int, Any]],
                                  atoms_frequencies: Dict[int, int]) -> Iterator[Dict[int, int]]:
        if len(atoms) == len(set(atoms.values())):
            return  # all atoms unique

        components, closures = cls.__compile_query(atoms, bonds, atoms_frequencies)
        groups = {x: n for n, x in enumerate(atoms)}
        mappers = [cls.__get_mapping(order, closures, atoms, bonds, {x for x, *_ in order}, groups)
                   for order in components]
        if len(mappers) == 1:
            for mapping in mappers[0]:
                if any(k != v for k, v in mapping.items()):
                    yield mapping
        for match in lazy_product(*mappers):
            mapping = match[0]
            for m in match[1:]:
                mapping.update(m)
            if any(k != v for k, v in mapping.items()):
                yield mapping


__all__ = ['Isomorphism']



# -*- coding: utf-8 -*-
#
#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import List, Union, Tuple, Dict
from . import cgr, molecule, query  # cyclic imports resolve
from .bonds import Bond, DynamicBond
from .common import Graph
from ..algorithms.calculate2d import Calculate2DCGR
from ..algorithms.depict import DepictQueryCGR
from ..algorithms.smiles import QueryCGRSmiles
from ..periodictable import (Element, DynamicElement, QueryElement, DynamicQueryElement, AnyElement, DynamicAnyElement,
                             AnyAtom)


class QueryCGRContainer(Graph, QueryCGRSmiles, DepictQueryCGR, Calculate2DCGR):
    __slots__ = ('_p_charges', '_p_radicals', '_neighbors', '_hybridizations', '_p_neighbors', '_p_hybridizations')

    def __init__(self):
        self._p_charges: Dict[int, int] = {}
        self._p_radicals: Dict[int, bool] = {}
        self._neighbors: Dict[int, Tuple[int, ...]] = {}
        self._hybridizations: Dict[int, Tuple[int, ...]] = {}
        self._p_neighbors: Dict[int, Tuple[int, ...]] = {}
        self._p_hybridizations: Dict[int, Tuple[int, ...]] = {}
        super().__init__()

    def add_atom(self, atom: Union[AnyAtom, int, str], *args,
                 p_charge: int = 0, p_is_radical: bool = False,
                 neighbors: Union[int, List[int], Tuple[int, ...], None] = None,
                 hybridization: Union[int, List[int], Tuple[int, ...], None] = None,
                 p_neighbors: Union[int, List[int], Tuple[int, ...], None] = None,
                 p_hybridization: Union[int, List[int], Tuple[int, ...], None] = None, **kwargs):
        neighbors = self._validate_neighbors(neighbors)
        p_neighbors = self._validate_neighbors(p_neighbors)
        hybridization = self._validate_hybridization(hybridization)
        p_hybridization = self._validate_hybridization(p_hybridization)
        neighbors, p_neighbors = self._validate_neighbors_pairing(neighbors, p_neighbors)
        hybridization, p_hybridization = self._validate_hybridization_pairing(hybridization, p_hybridization)

        p_charge = self._validate_charge(p_charge)
        p_is_radical = self._validate_radical(p_is_radical)

        if not isinstance(atom, (DynamicQueryElement, DynamicAnyElement)):
            if isinstance(atom, AnyElement):
                atom = DynamicAnyElement()
            elif isinstance(atom, (Element, QueryElement, DynamicElement)):
                atom = DynamicQueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
            elif isinstance(atom, str):
                atom = DynamicQueryElement.from_symbol(atom)()
            elif isinstance(atom, int):
                atom = DynamicQueryElement.from_atomic_number(atom)()
            else:
                raise TypeError('QueryElement object expected')

        _map = super().add_atom(atom, *args, **kwargs)
        self._p_charges[_map] = p_charge
        self._p_radicals[_map] = p_is_radical
        self._neighbors[_map] = neighbors
        self._hybridizations[_map] = hybridization
        self._p_neighbors[_map] = p_neighbors
        self._p_hybridizations[_map] = p_hybridization
        return _map

    def add_bond(self, n, m, bond: Union[DynamicBond, Bond, int]):
        if not isinstance(bond, DynamicBond):
            if isinstance(bond, Bond):
                bond = object.__new__(DynamicBond)
                bond._DynamicBond__order = bond._DynamicBond__p_order = bond.order
            else:
                bond = DynamicBond(bond, bond)
        super().add_bond(n, m, bond)

    def delete_atom(self, n):
        super().delete_atom(n)
        del self._p_charges[n]
        del self._p_radicals[n]
        del self._neighbors[n]
        del self._hybridizations[n]
        del self._p_neighbors[n]
        del self._p_hybridizations[n]

    def remap(self, mapping, *, copy=False) -> 'QueryCGRContainer':
        h = super().remap(mapping, copy=copy)
        mg = mapping.get
        spr = self._p_radicals
        sn = self._neighbors
        sh = self._hybridizations
        spn = self._p_neighbors
        sph = self._p_hybridizations

        if copy:
            hpc = h._p_charges
            hpr = h._p_radicals
            hn = h._neighbors
            hh = h._hybridizations
            hpn = h._p_neighbors
            hph = h._p_hybridizations
        else:
            hpc = {}
            hpr = {}
            hn = {}
            hh = {}
            hpn = {}
            hph = {}

        for n, c in self._p_charges.items():
            m = mg(n, n)
            hpc[m] = c
            hpr[m] = spr[n]
            hn[m] = sn[n]
            hpn[m] = spn[n]
            hh[m] = sh[n]
            hph[m] = sph[n]

        if copy:
            return h

        self._p_charges = hpc
        self._p_radicals = hpr
        self._neighbors = hn
        self._hybridizations = hh
        self._p_neighbors = hpn
        self._p_hybridizations = hph
        return self

    def copy(self, **kwargs) -> 'QueryCGRContainer':
        copy = super().copy(**kwargs)
        copy._neighbors = self._neighbors.copy()
        copy._hybridizations = self._hybridizations.copy()
        copy._p_neighbors = self._p_neighbors.copy()
        copy._p_hybridizations = self._p_hybridizations.copy()
        copy._p_radicals = self._p_radicals.copy()
        copy._p_charges = self._p_charges.copy()
        return copy

    def substructure(self, atoms, **kwargs) -> 'QueryCGRContainer':
        """
        create substructure containing atoms from atoms list

        :param atoms: list of atoms numbers of substructure
        :param meta: if True metadata will be copied to substructure
        """
        sub, atoms = super().substructure(atoms, self.__class__, **kwargs)
        sa = self._atoms
        spc = self._p_charges
        spr = self._p_radicals
        sn = self._neighbors
        sh = self._hybridizations
        spn = self._p_neighbors
        sph = self._p_hybridizations

        sub._p_charges = {n: spc[n] for n in atoms}
        sub._p_radicals = {n: spr[n] for n in atoms}
        sub._neighbors = {n: sn[n] for n in atoms}
        sub._hybridizations = {n: sh[n] for n in atoms}
        sub._p_neighbors = {n: spn[n] for n in atoms}
        sub._p_hybridizations = {n: sph[n] for n in atoms}

        sub._atoms = ca = {}
        for n in atoms:
            atom = sa[n].copy()
            ca[n] = atom
            atom._attach_to_graph(sub, n)
        return sub

    def union(self, other, **kwargs) -> 'QueryCGRContainer':
        if isinstance(other, (QueryCGRContainer, cgr.CGRContainer)):
            u, other = super().union(other, **kwargs)
            u._p_charges.update(other._p_charges)
            u._p_radicals.update(other._p_radicals)

            if isinstance(other, QueryCGRContainer):
                u._neighbors.update(other._neighbors)
                u._hybridizations.update(other._hybridizations)
                u._p_neighbors.update(other._p_neighbors)
                u._p_hybridizations.update(other._p_hybridizations)

                ua = u._atoms
                for n, atom in other._atoms.items():
                    atom = atom.copy()
                    ua[n] = atom
                    atom._attach_to_graph(u, n)
            else:  # CGRContainer
                un = u._neighbors
                uh = u._hybridizations
                upn = u._p_neighbors
                uph = u._p_hybridizations
                oh = other._hybridizations
                oph = other._p_hybridizations
                for n, m_bond in other._bonds.items():
                    un[n] = (sum(b.order is not None for b in m_bond.values()),)
                    upn[n] = (sum(b.p_order is not None for b in m_bond.values()),)
                    uh[n] = (oh[n],)
                    uph[n] = (oph[n],)

                ua = u._atoms
                for n, atom in other._atoms.items():
                    atom = DynamicQueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
                    ua[n] = atom
                    atom._attach_to_graph(u, n)

            ub = u._bonds
            for n in other._bonds:
                ub[n] = {}
            seen = set()
            for n, m_bond in other._bonds.items():
                seen.add(n)
                for m, bond in m_bond.items():
                    if m not in seen:
                        ub[n][m] = ub[m][n] = bond.copy()
            return u
        elif isinstance(other, (query.QueryContainer, molecule.MoleculeContainer)):
            u, other = super().union(other, **kwargs)
            u._p_charges.update(other._charges)
            u._p_radicals.update(other._radicals)

            if isinstance(other, query.QueryContainer):
                u._neighbors.update(other._neighbors)
                u._hybridizations.update(other._hybridizations)
                u._p_neighbors.update(other._neighbors)
                u._p_hybridizations.update(other._hybridizations)
            else:  # MoleculeContainer
                un = u._neighbors
                uh = u._hybridizations
                upn = u._p_neighbors
                uph = u._p_hybridizations
                oh = other._hybridizations
                for n, m_bond in other._bonds.items():
                    un[n] = upn[n] = (len(m_bond),)
                    uh[n] = uph[n] = (oh[n],)

            ua = u._atoms
            for n, atom in other._atoms.items():
                atom = DynamicQueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
                ua[n] = atom
                atom._attach_to_graph(u, n)

            ub = u._bonds
            for n in other._bonds:
                ub[n] = {}
            seen = set()
            for n, m_bond in other._bonds.items():
                seen.add(n)
                for m, bond in m_bond.items():
                    if m not in seen:
                        ub[n][m] = ub[m][n] = bc = object.__new__(DynamicBond)
                        bc._DynamicBond__order = bc._DynamicBond__p_order = bond.order
            return u
        else:
            raise TypeError('Graph expected')

    def get_mapping(self, other: Union['QueryCGRContainer', 'cgr.CGRContainer'], **kwargs):
        if isinstance(other, (QueryCGRContainer, cgr.CGRContainer)):
            return super().get_mapping(other, **kwargs)
        raise TypeError('CGRContainer or QueryCGRContainer expected')

    def get_mcs_mapping(self, other: Union['QueryCGRContainer', 'cgr.CGRContainer'], **kwargs):
        if isinstance(other, (QueryCGRContainer, cgr.CGRContainer)):
            return super().get_mcs_mapping(other, **kwargs)
        raise TypeError('CGRContainer or QueryCGRContainer expected')

    @staticmethod
    def _validate_neighbors(neighbors):
        if neighbors is None:
            neighbors = ()
        elif isinstance(neighbors, int):
            if neighbors < 0 or neighbors > 14:
                raise ValueError('neighbors should be in range [0, 14]')
            neighbors = (neighbors,)
        elif isinstance(neighbors, (tuple, list)):
            if not all(isinstance(n, int) for n in neighbors):
                raise TypeError('neighbors should be list or tuple of ints')
            if any(n < 0 or n > 14 for n in neighbors):
                raise ValueError('neighbors should be in range [0, 14]')
            neighbors = tuple(neighbors)
        else:
            raise TypeError('neighbors should be int or list or tuple of ints')
        return neighbors

    @staticmethod
    def _validate_hybridization(hybridization):
        if hybridization is None:
            hybridization = ()
        elif isinstance(hybridization, int):
            if hybridization < 1 or hybridization > 4:
                raise ValueError('hybridization should be in range [1, 4]')
            hybridization = (hybridization,)
        elif isinstance(hybridization, (tuple, list)):
            if not all(isinstance(h, int) for h in hybridization):
                raise TypeError('hybridizations should be list or tuple of ints')
            if any(h < 1 or h > 4 for h in hybridization):
                raise ValueError('hybridizations should be in range [1, 4]')
            hybridization = tuple(hybridization)
        else:
            raise TypeError('hybridization should be int or list or tuple of ints')
        return hybridization

    @staticmethod
    def _validate_neighbors_pairing(neighbors, p_neighbors):
        if len(neighbors) != len(p_neighbors):
            raise ValueError('neighbors and p_neighbors should be same length')
        if neighbors:
            if len(set(zip(neighbors, p_neighbors))) != len(neighbors):
                raise ValueError('paired neighbors and p_neighbors should be unique')
            neighbors, p_neighbors = zip(*sorted(zip(neighbors, p_neighbors)))
        return neighbors, p_neighbors

    @staticmethod
    def _validate_hybridization_pairing(hybridization, p_hybridization):
        if len(hybridization) != len(p_hybridization):
            raise ValueError('hybridization and p_hybridization should be same length')
        if hybridization:
            if len(set(zip(hybridization, p_hybridization))) != len(hybridization):
                raise ValueError('paired hybridization and p_hybridization should be unique')
            hybridization, p_hybridization = zip(*sorted(zip(hybridization, p_hybridization)))
        return hybridization, p_hybridization

    def __getstate__(self):
        return {'p_charges': self._p_charges, 'p_radicals': self._p_radicals, 'neighbors': self._neighbors,
                'hybridizations': self._hybridizations, 'p_neighbors': self._p_neighbors,
                'p_hybridizations': self._p_hybridizations, **super().__getstate__()}

    def __setstate__(self, state):
        super().__setstate__(state)
        self._neighbors = state['neighbors']
        self._hybridizations = state['hybridizations']
        self._p_neighbors = state['p_neighbors']
        self._p_hybridizations = state['p_hybridizations']
        self._p_charges = state['p_charges']
        self._p_radicals = state['p_radicals']


__all__ = ['QueryCGRContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod
from CachedMethods import cached_property, cached_args_method
from typing import Dict, Optional, Tuple, Iterable, Iterator, Union, List, Type
from .bonds import Bond, DynamicBond
from ..algorithms.components import GraphComponents
from ..algorithms.isomorphism import Isomorphism
from ..algorithms.mcs import MCS
from ..algorithms.morgan import Morgan
from ..algorithms.sssr import SSSR
from ..exceptions import AtomNotFound
from ..periodictable import AnyAtom


class Graph(GraphComponents, Morgan, SSSR, Isomorphism, MCS, ABC):
    __slots__ = ('_atoms', '_bonds', '_plane', '_charges', '_radicals', '__meta', '__name', '_parsed_mapping',
                 '__dict__', '__weakref__')

    def __init__(self):
        """
        Empty data object initialization or conversion from another object type
        """
        self._atoms: Dict[int, AnyAtom] = {}
        self._charges: Dict[int, int] = {}
        self._radicals: Dict[int, bool] = {}
        self._plane: Dict[int, Tuple[float, float]] = {}
        self._bonds: Dict[int, Dict[int, Union[Bond, DynamicBond]]] = {}
        self._parsed_mapping: Dict[int, int] = {}
        self.__meta = {}
        self.__name = ''

    def __getstate__(self):
        return {'atoms': self._atoms, 'bonds': self._bonds, 'meta': self.__meta, 'plane': self._plane,
                'parsed_mapping': self._parsed_mapping, 'charges': self._charges, 'radicals': self._radicals,
                'name': self.__name}

    def __setstate__(self, state):
        self._atoms = state['atoms']
        for n, a in state['atoms'].items():
            a._attach_to_graph(self, n)
        self._charges = state['charges']
        self._radicals = state['radicals']
        self._plane = state['plane']
        self._bonds = state['bonds']
        self._parsed_mapping = state['parsed_mapping']
        self.__meta = state['meta']
        self.__name = state.get('name', '')  # 4.0.9 compatibility

    def __len__(self):
        return len(self._atoms)

    def __iter__(self):
        return iter(self._atoms)

    def __contains__(self, n: int):
        return n in self._atoms

    def __bool__(self):
        return bool(self._atoms)

    def atom(self, n: int) -> AnyAtom:
        return self._atoms[n]

    def has_atom(self, n: int) -> bool:
        return n in self._atoms

    def atoms(self) -> Iterator[Tuple[int, AnyAtom]]:
        """
        iterate over all atoms
        """
        return iter(self._atoms.items())

    @cached_property
    def atoms_count(self) -> int:
        return len(self._atoms)

    @cached_property
    def atoms_numbers(self) -> Tuple[int, ...]:
        return tuple(self._atoms)

    @cached_args_method
    def environment(self, atom: int) -> Tuple[Tuple[Union[Bond, DynamicBond], AnyAtom], ...]:
        """
        pairs of (bond, atom) connected to atom

        :param atom: number
        """
        atoms = self._atoms
        return tuple((bond, atoms[n]) for n, bond in self._bonds[atom].items())

    def bond(self, n: int, m: int) -> Union[Bond, DynamicBond]:
        return self._bonds[n][m]

    def has_bond(self, n: int, m: int) -> bool:
        try:
            self._bonds[n]  # check if atom exists
            return n in self._bonds[m]
        except KeyError:
            raise AtomNotFound

    def bonds(self) -> Iterator[Tuple[int, int, Union[Bond, DynamicBond]]]:
        """
        iterate other all bonds
        """
        seen = set()
        for n, m_bond in self._bonds.items():
            seen.add(n)
            for m, bond in m_bond.items():
                if m not in seen:
                    yield n, m, bond

    @cached_property
    def bonds_count(self) -> int:
        return sum(len(x) for x in self._bonds.values()) // 2

    @property
    def meta(self) -> Dict:
        return self.__meta

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError('name should be string up to 80 symbols')
        if len(name) > 80:
            raise ValueError('name should be string up to 80 symbols')
        self.__name = name

    @abstractmethod
    def add_atom(self, atom, _map: Optional[int] = None, *, charge: int = 0,
                 is_radical: bool = False, xy: Tuple[float, float] = (0., 0.)) -> int:
        """
        new atom addition
        """
        if _map is None:
            _map = max(self._atoms, default=0) + 1
        elif not isinstance(_map, int):
            raise TypeError('mapping should be integer')
        elif _map in self._atoms:
            raise ValueError('atom with same number exists')

        if not isinstance(xy, tuple) or len(xy) != 2 or not isinstance(xy[0], float) or not isinstance(xy[1], float):
            raise TypeError('XY should be tuple with 2 float')

        charge = self._validate_charge(charge)
        is_radical = self._validate_radical(is_radical)

        self._atoms[_map] = atom
        self._charges[_map] = charge
        self._radicals[_map] = is_radical
        self._plane[_map] = xy
        self._bonds[_map] = {}
        atom._attach_to_graph(self, _map)
        self.__dict__.clear()
        return _map

    @abstractmethod
    def add_bond(self, n: int, m: int, bond):
        """
        new bond addition
        """
        if n == m:
            raise ValueError('atom loops impossible')
        if n not in self._bonds or m not in self._bonds:
            raise AtomNotFound('atoms not found')
        if n in self._bonds[m]:
            raise ValueError('atoms already bonded')

        self._bonds[n][m] = self._bonds[m][n] = bond
        self.__dict__.clear()

    @abstractmethod
    def delete_atom(self, n: int):
        """
        implementation of atom removing
        """
        del self._atoms[n]
        del self._charges[n]
        del self._radicals[n]
        del self._plane[n]
        sb = self._bonds
        for m in sb.pop(n):
            del sb[m][n]
        try:
            del self._parsed_mapping[n]
        except KeyError:
            pass
        self.__dict__.clear()

    def delete_bond(self, n: int, m: int):
        """
        implementation of bond removing
        """
        del self._bonds[n][m]
        del self._bonds[m][n]
        self.__dict__.clear()

    @abstractmethod
    def remap(self, mapping: Dict[int, int], *, copy: bool = False):
        if len(mapping) != len(set(mapping.values())):
            raise ValueError('mapping overlap')

        mg = mapping.get
        sp = self._plane
        sc = self._charges
        sr = self._radicals

        if copy:
            h = self.__class__()
            h._Graph__meta.update(self.__meta)
            h._Graph__name = self.__name
            hb = h._bonds
            ha = h._atoms
            hc = h._charges
            hr = h._radicals
            hp = h._plane
            hm = h._parsed_mapping
            for n, atom in self._atoms.items():
                m = mg(n, n)
                hc[m] = sc[n]
                hr[m] = sr[n]
                hp[m] = sp[n]
                atom = atom.copy()
                ha[m] = atom
                atom._attach_to_graph(h, m)
        else:
            hb = {}
            ha = {}
            hc = {}
            hr = {}
            hp = {}
            hm = {}
            for n, atom in self._atoms.items():
                m = mg(n, n)
                hc[m] = sc[n]
                hr[m] = sr[n]
                hp[m] = sp[n]
                ha[m] = atom
                atom._change_map(m)  # change mapping number
            self._atoms = ha
            self._charges = hc
            self._radicals = hr
            self._plane = hp

        for n, m_bond in self._bonds.items():
            hb[mg(n, n)] = {mg(m, m): b for m, b in m_bond.items()}

        for n, m in self._parsed_mapping.items():
            hm[mg(n, n)] = m

        if copy:
            return h

        self._bonds = hb
        self._parsed_mapping = hm
        self.__dict__.clear()
        return self

    @abstractmethod
    def copy(self, *, meta: bool = True):
        """
        copy of graph

        :param meta: include metadata
        """
        copy = object.__new__(self.__class__)
        if meta:
            copy._Graph__meta = self.__meta.copy()
            copy._Graph__name = self.__name
        else:
            copy._Graph__meta = {}
            copy._Graph__name = ''

        copy._charges = self._charges.copy()
        copy._radicals = self._radicals.copy()
        copy._plane = self._plane.copy()
        copy._parsed_mapping = self._parsed_mapping.copy()

        copy._bonds = cb = {}
        for n, m_bond in self._bonds.items():
            cb[n] = cbn = {}
            for m, bond in m_bond.items():
                if m in cb:  # bond partially exists. need back-connection.
                    cbn[m] = cb[m][n]
                else:
                    cbn[m] = bond.copy()

        copy._atoms = ca = {}
        for n, atom in self._atoms.items():
            atom = atom.copy()
            ca[n] = atom
            atom._attach_to_graph(copy, n)
        return copy

    @abstractmethod
    def substructure(self, atoms: Iterable[int], sub: Type['Graph'], *, meta: bool = False):
        if not atoms:
            raise ValueError('empty atoms list not allowed')
        if set(atoms) - self._atoms.keys():
            raise ValueError('invalid atom numbers')
        atoms = tuple(n for n in self._atoms if n in atoms)  # save original order
        sub = object.__new__(sub)

        sc = self._charges
        sr = self._radicals
        sp = self._plane
        sb = self._bonds

        if meta:
            sub._Graph__meta = self.__meta.copy()
            sub._Graph__name = self.__name
        else:
            sub._Graph__meta = {}
            sub._Graph__name = ''

        sub._charges = {n: sc[n] for n in atoms}
        sub._radicals = {n: sr[n] for n in atoms}
        sub._plane = {n: sp[n] for n in atoms}
        sub._parsed_mapping = {n: m for n, m in self._parsed_mapping.items() if n in atoms}

        sub._bonds = cb = {}
        for n in atoms:
            cb[n] = cbn = {}
            for m, bond in sb[n].items():
                if m in cb:  # bond partially exists. need back-connection.
                    cbn[m] = cb[m][n]
                elif m in atoms:
                    cbn[m] = bond.copy()
        return sub, atoms

    def __and__(self, other):
        """
        Substructure of graph with given nodes.
        """
        return self.substructure(other)

    def __sub__(self, other):
        """
        Given nodes excluded substructure of graph.
        """
        atoms = set(other)
        if atoms - self._atoms.keys():
            raise ValueError('invalid atom numbers')
        atoms = self._atoms.keys() - atoms
        if atoms:
            return self.substructure(atoms)
        raise ValueError('full substitution not allowed')

    def augmented_substructure(self, atoms: Iterable[int], deep: int = 1, **kwargs) -> 'Graph':
        """
        create substructure containing atoms and their neighbors

        :param atoms: list of core atoms in graph
        :param deep: number of bonds between atoms and neighbors
        :param meta: copy metadata to each substructure
        :param as_query: return Query object based on graph substructure. for Molecule and CGR only
        """
        return self.substructure(self._augmented_substructure(atoms, deep)[-1], **kwargs)

    def augmented_substructures(self, atoms: Iterable[int], deep: int = 1, **kwargs) -> List['Graph']:
        """
        create list of substructures containing atoms and their neighbors

        :param atoms: list of core atoms in graph
        :param deep: number of bonds between atoms and neighbors
        :param meta: copy metadata to each substructure
        :param as_query: return Query object based on graph substructure. for Molecule and CGR only
        :return: list of graphs containing atoms, atoms + first circle, atoms + 1st + 2nd,
            etc up to deep or while new nodes available
        """
        return [self.substructure(a, **kwargs) for a in self._augmented_substructure(atoms, deep)]

    def union(self, other: 'Graph', *, remap=False) -> 'Graph':
        """
        Merge Graphs into one.

        :param remap: if atoms has collisions then remap other graph atoms else raise exception.
        """
        if self._atoms.keys() & other._atoms.keys():
            if remap:
                other = other.remap({n: i for i, n in enumerate(other, start=max(self._atoms) + 1)}, copy=True)
            else:
                raise ValueError('mapping of graphs is not disjoint')

        u = self.copy(meta=False)
        u._charges.update(other._charges)
        u._radicals.update(other._radicals)
        u._plane.update(other._plane)
        u._parsed_mapping.update(other._parsed_mapping)
        return u, other

    def __or__(self, other):
        """
        G | H is union of graphs
        """
        return self.union(other)

    def split(self, meta: bool = False) -> List['Graph']:
        """
        split disconnected structure to connected substructures

        :param meta: copy metadata to each substructure
        :return: list of substructures
        """
        return [self.substructure(c, meta=meta) for c in self.connected_components]

    def flush_cache(self):
        self.__dict__.clear()

    @staticmethod
    def _validate_charge(charge):
        if not isinstance(charge, int):
            raise TypeError('formal charge should be int in range [-4, 4]')
        if charge > 4 or charge < -4:
            raise ValueError('formal charge should be in range [-4, 4]')
        return charge

    @staticmethod
    def _validate_radical(is_radical):
        if not isinstance(is_radical, bool):
            raise TypeError('radical state should be bool')
        return is_radical


__all__ = ['Graph']



# -*- coding: utf-8 -*-
#
#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Tagir Akhmetshin <tagirshin@gmail.com>
#  Copyright 2019 Alexander Nikanshin <17071996sasha@gmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import FrozenDict
from .element import Element
from .groups import GroupVII
from .periods import PeriodIV, PeriodV, PeriodVI, PeriodVII


class Mn(Element, PeriodIV, GroupVII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 25

    @property
    def isotopes_distribution(self):
        return FrozenDict({55: 1.0, 52: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({55: 54.93805, 52: 51.945566})

    @property
    def _common_valences(self):
        return 0, 2

    @property
    def _valences_exceptions(self):
        return ((2, False, 0, ()), (3, False, 0, ()),
                (0, False, 0, ((2, 'O'), (2, 'O'))),
                (0, False, 0, ((2, 'O'),)),  # MnO
                (0, False, 0, ((2, 'O'), (1, 'O'))),  # Mn2O3
                (0, False, 0, ((2, 'O'), (2, 'O'))),  # MnO2
                (0, False, 0, ((2, 'O'), (2, 'O'), (1, 'O'), (1, 'O'))),  # [MnO4]2-
                (0, False, 0, ((2, 'O'), (2, 'O'), (2, 'O'), (1, 'O'))))  # [MnO4]-

    @property
    def atomic_radius(self):
        return 1.61


class Tc(Element, PeriodV, GroupVII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 43

    @property
    def isotopes_distribution(self):
        return FrozenDict({99: 1.0, 98: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({99: 98.906255, 98: 97.907216})

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ((0, False, 0, ((2, 'O'), (2, 'O'))),  # TcO2
                (0, False, 0, ((2, 'S'), (2, 'S'))))  # TcS2

    @property
    def atomic_radius(self):
        return 1.83


class Re(Element, PeriodVI, GroupVII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 75

    @property
    def isotopes_distribution(self):
        return FrozenDict({185: 0.374, 187: 0.626, 186: 0., 188: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({185: 184.952956, 187: 186.955751, 186: 185.954986, 188: 187.958114})

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ()

    @property
    def atomic_radius(self):
        return 1.88


class Bh(Element, PeriodVII, GroupVII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 107

    @property
    def isotopes_distribution(self):
        return FrozenDict({270: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({270: 270.133363})

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ()

    @property
    def atomic_radius(self):
        return 1.88  # unknown, taken radius of previous element in group


__all__ = ['Mn', 'Tc', 'Re', 'Bh']
